#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom Qt tab widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, Property, QPoint, QPointF, QRect, QSize
from Qt.QtWidgets import QWidget, QTabWidget, QMainWindow
from Qt.QtGui import QColor, QPalette

from tpDcc import dcc
from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.core import qtutils
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, window, tabbars, buttons, group, dividers, stack


class BaseTabWidget(QTabWidget, object):
    """
    Base class for tab widgets
    """

    deletedTab = Signal(object)

    def __init__(self, parent=None):
        super(BaseTabWidget, self).__init__(parent=parent)

        self._set_custom_tab_bar()
        self.tabCloseRequested.connect(self.close_tab)

        self._show_close_message = False
        self._show_text = ''
        self._show_title = ''
        self._show_fn = None
        self._show_can_cancel = False

    def removeTab(self, index):
        """
        Function that is called when a tab is removed
        :param index: int, index of the removed tab
        """

        if self._show_close_message:
            can_close = qtutils.get_permission(
                message='Save before closing tab?',
                cancel=self._show_can_cancel, title='Closing tab', parent=self)
            if can_close is None:
                return
            elif can_close:
                if self._show_fn:
                    self._show_fn(index, self.widget(index))

        super(BaseTabWidget, self).removeTab(index)
        self.deletedTab.emit(index)

    def close_tab(self, index):
        """
        Function that is called before
        :param index:
        :return:
        """

        return self.removeTab(index)

    def set_close_message(self, message_text, message_title, cancel=False, accept_fn_call=None):
        self._show_close_message = True
        self._show_text = message_text
        self._show_title = message_title
        self._show_can_cancel = cancel
        self._show_fn = accept_fn_call

    def get_unique_tab_name(self, name, curr_index=0):
        """
        Returns a unique tab name for the current tab widget
        :param name: str, base name to get a unique name
        :param curr_index: int, starting index to start searching for unique names
        :return:
        """

        curr_index = curr_index
        new_name = name
        index = 0
        if curr_index > 0:
            new_name = name + '_' + str(curr_index)
        for i in range(self.tabBar().count()):
            curr_tab_name = self.tabBar().tabText(i)
            if curr_tab_name == new_name:
                index = index + 1
        if index > 0:
            return self.get_unique_tab_name(name, curr_index=curr_index + index)
        return new_name

    def update_tab_bar_signals(self):
        """
        Override functionality if necessary
        Updates teh tab bar signals
        """

        pass

    def _set_custom_tab_bar(self):
        """
        Sets a custom tab bar with editable functionality for the widget
        """

        pass


class BaseEditableTabWidget(BaseTabWidget, object):
    """
    Tab widget with basic editable functionality
    """

    def __init__(self, parent=None):
        super(BaseEditableTabWidget, self).__init__(parent=parent)

        self.setMovable(True)
        self.setTabsClosable(True)
        self._edit_mode = True

    def set_edit_mode(self, value):
        if self._edit_mode != value:
            self._edit_mode = value
            self.setMovable(value)
            self.setTabsClosable(value)

    def get_edit_mode(self):
        return self._edit_mode

    edit_mode = property(get_edit_mode, set_edit_mode)

    def mousePressEvent(self, event):
        pos = event.pos()
        rect = self._add_tab_btn_rect()
        if rect.contains(pos):
            pass
        # super(BaseEditableTabWidget, self).mousePressEvent(event)

    def _add_tab_btn_rect(self):
        """
        Returns the visual rect of the current selected tab
        :return: QRect
        """

        return self.tabBar().add_tab_btn.rect()

    def _set_custom_tab_bar(self):
        """
        Sets a custom tab bar with editable functionality for the widget
        """

        self.setTabBar(tabbars.EditableTabBar(self))
        self.update_tab_bar_signals()


class TearOffTabWidget(BaseTabWidget, object):
    """
    Tab widget with basic tear off functionality
    """

    class TabWidgetStorage(object):
        def __init__(self, index, widget, title, visible=True):
            self.index = index
            self.widget = widget
            self.title = title
            self.visible = visible
            self.detached = None

        def set_detached(self, detached):
            self.detached = detached

        def __repr__(self):
            return 'index %d, widget %r, title %s, visible %r, detached %r' % (
                self.index, self.widget, self.title, self.visible, self.detached)

    tabAdded = Signal()
    tabDetached = Signal(int, QPoint)
    tabBarRenamed = Signal(str, str)
    requestRemove = Signal(int)
    tabChanged = Signal(int)
    undoOpen = Signal()
    undoClose = Signal()

    def __init__(self, parent=None):
        super(TearOffTabWidget, self).__init__(parent=parent)

        self._previous_tab_id = 0
        self.tab_idx = dict()

        self.setMovable(False)
        self.setTabsClosable(True)
        self.setMouseTracking(True)
        self.setObjectName('BaseTabWidget')

        self._set_palette_color()

        # don't draw a frame around the tab contents
        self.setStyleSheet('QTabWidget:tab:disabled{width:0;height:0;margin:0;padding:0;border:none}')
        self.setDocumentMode(True)

    def update_tab_bar_signals(self):
        """
        Updates the signals attached to the tab bar
        """

        tab_bar = self.tabBar()
        if tab_bar is None:
            return
        self.tabBar().tabDetached.connect(self.tabDetached.emit)
        self.tabBar().tabDetached.connect(self._on_detach_tab)
        self.tabBar().tabMoved.connect(self._on_move_tab)
        self.currentChanged.connect(self._on_tab_changed_tab)

    def addTab(self, *args, **kwargs):
        added_tab_index = super(TearOffTabWidget, self).addTab(*args, **kwargs)
        self.tab_inserted(added_tab_index)

    def removeTab(self, index):
        super(TearOffTabWidget, self).removeTab(index)
        self.tab_idx.pop(index)

    def tab_inserted(self, index):
        w = self.widget(index)
        for i in self.tab_idx.values():
            if i.widget == w:
                return
        self.tab_idx[index] = self.TabWidgetStorage(index, self.widget(index), self.tabText(index))

    def set_widget_visible(self, widget, visible):
        w = self._find_first_window(widget)
        for i in self.tab_idx.values():
            if i.widget == w:
                if visible:
                    if not i.visible:
                        new_index = -1
                        for j in self.tab_idx.values():
                            if j.visible and j.index > i.index:
                                c_idx = self._tab_widget_index(j.widget)
                                if c_idx < i.index and c_idx != -1:
                                    new_index = c_idx
                                else:
                                    new_index = i.index
                                break
                        self.insertTab(new_index, i.widget, i.title)
                else:
                    i.visible = False
                    index = self._tab_widget_index(i.widget)
                    self.removeTab(index)

    def add_panel(self, widget, label):

        # TODO: We should not storing the detach config here
        # sg_group = window.DetachedWindow.SettingGroup(label)
        # with sg_group as settings:
        #     detached = settings.value('detached', False)
        detached = False

        index = len(self.tab_idx)
        self.tab_idx[index] = self.TabWidgetStorage(index, widget, label)
        if not detached:
            index = self.addTab(widget, label)
            for i in self.tab_idx.values():
                if i.widget == widget:
                    i.set_detached(None)
        else:
            detach_window = window.DetachedWindow(label, self.parentWidget())
            detach_window.tab_idx = index
            detach_window.setAttribute(Qt.WA_DeleteOnClose, True)
            self.tab_idx[index].set_detached(detach_window)
            detach_window.windowClosed.connect(self._on_attach_tab)

            panel = self._get_widget(widget)
            panel.widgetVisible.disconnect(self.set_widget_visible)
            panel.widgetVisible.connect(detach_window.set_widget_visible)
            widget.setParent(detach_window)

            detach_window.set_widget(widget)
            detach_window.destroyed.connect(detach_window.deleteLater)
            # detach_window.restoreGeometry(settings.value('geometry', ''))
            detach_window.show()

    def clear(self):
        """
        Delete all the tabs added to this tab widget
        """

        all_tabs = [self.widget(i) for i in range(self.count())]
        self.blockSignals(True)
        for tab in all_tabs:
            tab.deleteLater()
        super(TearOffTabWidget, self).clear()
        self.blockSignals(False)

    def dragEnterEvent(self, event):
        """
        Override drag enter behaviour
        """

        event.accept()
        super(TearOffTabWidget, self).dragEnterEvent(event)

    def dragMoveEvent(self, event):
        """
        Override drag move behaviour
        """

        event.accept()
        super(TearOffTabWidget, self).dragMoveEvent(event)
    # endregion

    # region Private Functions
    def _set_palette_color(self):
        """
        Sets the initial for the different elements of the tab widget
        Override to create custom color styles for this widget
        """

        palette = self.palette()
        self.border_color = QColor(Qt.black)
        self.button_color = palette.color(QPalette.Button)
        self.fill_color = palette.color(QPalette.Window)
        self.highlight_color = palette.color(QPalette.Highlight)

    def _set_custom_tab_bar(self):
        """
        Sets a custom tab bar with tear off functionality for the widget
        Override to add a custom tab bar for this widget
        """

        tab_bar = tabbars.TearOffTabBar()
        self.setTabBar(tab_bar)
        self.update_tab_bar_signals()

    def _find_first_window(self, w):
        widget = w
        while True:
            parent = widget.parent()
            if not parent:
                break
            widget = parent
            if isinstance(widget, QMainWindow):
                break

        return widget

    def _tab_widget_index(self, widget):
        for i in range(self.tabBar().count()):
            if self.widget(i) == widget:
                return i

        return -1

    def _get_widget(self, widget):
        panel = widget
        if isinstance(widget, QMainWindow):
            panel = widget.centralWidget()
            if panel.layout():
                panel = panel.layout().itemAt(0).widget()
        elif isinstance(widget, window.DetachedWindow.DetachPanel):
            if panel.layout():
                panel = panel.layout().itemAt(0).widget()

        return panel

    def _on_attach_tab(self, detach_window):
        detach_window.windowClosed.disconnect(self._on_attach_tab)
        detach_window.save_settings(False)
        panel = detach_window.centralWidget()
        tear_off_widget = self._get_widget(panel)
        try:
            panel.widgetVisible.disconnect(detach_window.set_widget_visible)
        except Exception:
            pass
        tear_off_widget.setParent(self)
        panel.widgetVisible.connect(self.set_widget_visible)

        new_index = -1

        for i in range(self.tabBar().count()):
            w = self.widget(i)
            for j in self.tab_idx.values():
                if j.widget == w and j.index > detach_window.tab_idx:
                    new_index = i
                    break
            else:
                continue
            break

        if new_index == -1:
            new_index = self.tabBar().count()

        new_index = self.insertTab(new_index, tear_off_widget, detach_window.windowTitle())

        if new_index != -1:
            self.setCurrentIndex(new_index)

        self.tab_idx[new_index].detached = None

    def _on_detach_tab(self, index, point):
        detach_window = window.DetachedWindow(self.tabText(index), self.parentWidget())
        w = self.widget(index)
        for i in self.tab_idx.values():
            if i.widget == w:
                detach_window.tab_idx = self.tab_idx[i.index].index
                self.tab_idx[i.index].detached = detach_window
                break

        detach_window.windowClosed.connect(self._on_attach_tab)

        tear_off_widget = self.widget(index)
        # panel = detach_window.centralWidget()
        # panel.widgetVisible.disconnect(self.set_widget_visible)
        # panel.widgetVisible.connect(detach_window.set_widget_visible)
        tear_off_widget.setParent(detach_window)

        if self.count() < 0:
            self.setCurrentIndex(0)

        detach_window.set_widget(tear_off_widget)
        detach_window.resize(tear_off_widget.size())
        detach_window.move(point)
        detach_window.show()

    def _on_move_tab(self, from_index, to_index):
        w = self.widget(from_index)
        text = self.tabText(from_index)
        self.removeTab(from_index)
        self.insertTab(to_index, w, text)
        self.setCurrentIndex(to_index)

    def _on_tab_changed_tab(self, index):
        self._previous_tab_id = index


class EditableTearOffTabWidget(TearOffTabWidget, object):
    """
    Custom tab widget with editable functionality
    """

    tab_size_changed = Signal(QSize)
    tab_label_renamed = Signal(str, str)
    tab_removed_text = Signal(str)
    window_modified = Signal()
    add_item_on = Signal(QPointF, QColor, int)
    add_item_with_set = Signal(QPointF, QColor, str, int)
    add_previewed_item = Signal(QPointF, QColor, str, int)
    copy_items = Signal(list)
    selected_items_changed = Signal(list)

    def __init__(self, parent=None):
        super(EditableTearOffTabWidget, self).__init__(parent=parent)

        self._is_editable = True
        self._has_highlight = False
        self.setAcceptDrops(True)
        self.setObjectName('EditableTabWidget')

    def set_is_editable(self, value):
        if self._is_editable != value:
            self._is_editable = value
            self.setMovable(value)
            self.update()

    def get_is_editable(self):
        return self._is_editable

    def set_has_highlight(self, value):
        if self._has_highlight != value:
            self._has_highlight = value
            self.update()

    def get_has_highlight(self):
        return self._has_highlight

    is_editable = property(get_is_editable, set_is_editable)
    has_highlight = property(get_has_highlight, set_has_highlight)

    def toggle_editable(self):
        """
        Toggles editable property of the tab widget
        """

        self.is_editable = not self.is_editable

    def delete_tab(self, index):
        """
        Deletes a tab with given index
        :param index: index of the tab to delete
        """

        if index < 0:
            return
        widget = self.widget(index)
        tab_text = self.tabText(index)
        super(EditableTearOffTabWidget, self).removeTab(index)
        widget.deleteLater()
        self.window_modified.emit()
        self.tab_removed_text.emit(tab_text)

    def button_rect(self):
        r = self.tabBar().tabRect(self.count() - 1)

        if dcc.is_maya() and dcc.get_version() > 2015:
            rect = QRect(2, 0, 30, 31)
        else:
            rect = QRect(6, 3, 26, 17)
        if r.isValid():
            if dcc.is_maya() and dcc.get_version() > 2015:
                rect.moveBottomLeft(r.bottomRight() + QPoint(1, 1))
            else:
                rect.moveBottomLeft(r.bottomRight() + QPoint(3, -1))
        return rect

    def update_tab_bar_signals(self):
        """
        Updates the signals attached to the tab bar
        """

        super(EditableTearOffTabWidget, self).update_tab_bar_signals()
        tab_bar = self.tabBar()
        if tab_bar is None:
            return
        tab_bar.tabMoved.connect(lambda: self.window_modified.emit())
        tab_bar.tab_label_renamed.connect(lambda: self.window_modified.emit())
        tab_bar.tab_label_renamed.connect(self.tab_label_renamed.emit)
        tab_bar.request_remove.connect(self.delete_tab)

    def _set_custom_tab_bar(self):
        """
        Sets a custom tab bar with editable and tear off functionality for the widget
        """

        tabBar = tabbars.EditableTearOffTabBar(self)
        self.setTabBar(tabBar)
        self.update_tab_bar_signals()
        self.setTabsClosable(True)


class MenuTabBlockButton(buttons.BaseToolButton, object):
    def __init__(self, parent=None):
        super(MenuTabBlockButton, self).__init__(parent=parent)

        self.setCheckable(True)


@theme.mixin
class MenuTabBlockButtonGroup(group.BaseButtonGroup, object):
    checkedChanged = Signal(int)

    def __init_(self, parent=None):
        super(MenuTabBlockButtonGroup, self).__init__(parent=parent)

        self.set_spacing(1)
        self._button_group.setExclusive(True)
        self._button_group.buttonClicked[int].connect(self.checkedChanged)

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_checked(self):
        """
        Returns current checked button's ID
        :return:  int
        """

        return self._button_group.checkedId()

    def _set_checked(self, value):
        """
        Sets current checked button's ID
        :param value: int
        """

        btn = self._button_group.button(value)
        btn.setChecked(True)
        self.checkedChanged.emit(value)

    checked = Property(int, _get_checked, _set_checked, notify=checkedChanged)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def create_button(self, data_dict):
        button_theme = self.theme()

        btn = MenuTabBlockButton()
        if data_dict.get('image'):
            btn.image(data_dict.get('image'))
        if data_dict.get('text'):
            if data_dict.get('image') or data_dict.get('icon'):
                btn.text_beside_icon()
            else:
                btn.text_only()
        btn.theme_size = button_theme.large if button_theme else theme.Theme.Sizes.LARGE

        return btn


class MenuTabWidget(base.BaseWidget, object):
    def __init__(self, parent=None):
        super(MenuTabWidget, self).__init__(parent=parent)

    def get_main_layout(self):
        main_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))

        return main_layout

    def ui(self):
        super(MenuTabWidget, self).ui()

        self.tool_btn_grp = MenuTabBlockButtonGroup()

        bar_widget = QWidget()
        bar_widget.setObjectName('bar_widget')
        self._bar_layout = layouts.HorizontalLayout(margins=(10, 0, 10, 0))
        bar_widget.setLayout(self._bar_layout)
        self._bar_layout.addWidget(self.tool_btn_grp)
        self._bar_layout.addStretch()

        self.main_layout.addWidget(bar_widget)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addSpacing(5)

    def append_widget(self, widget):
        """
        Adds a new widget to the end of the toolbar
        :param widget: QWidget
        """

        self._bar_layout.addWidget(widget)

    def insert_widget(self, widget):
        """
        Inserts the widget to the start of the toolbar
        :param widget: QWidget
        """

        self._bar_layout.insertWidget(0, widget)

    def add_menu(self, data_dict, index=None):
        """
        Adds a new menu
        :param data_dict: dict
        :param index: int
        """

        self.tool_btn_grp.add_button(data_dict, index)


class MenuLineButton(buttons.BaseToolButton, object):
    def __init__(self, parent=None):
        super(MenuLineButton, self).__init__(parent=parent)

        self.setCheckable(True)


class MenuLineButtonGroup(group.BaseButtonGroup, object):
    checkedChanged = Signal(int)

    def __init__(self, parent=None):
        super(MenuLineButtonGroup, self).__init__(parent=parent)

        self.set_spacing(1)
        self._button_group.setExclusive(True)
        self._button_group.buttonClicked[int].connect(self.checkedChanged)

    def create_button(self, data_dict):
        new_button = MenuLineButton(parent=self)
        if data_dict.get('image'):
            new_button.image(data_dict.get('image'))
        if data_dict.get('text'):
            if data_dict.get('image') or data_dict.get('icon'):
                new_button.text_beside_icon()
            else:
                new_button.text_only()

        return new_button

    def _get_checked(self):
        """
        Returns current checked button's ID
        :return: int
        """

        return self._button_group.checkedId()

    def _set_checked(self, value):
        """
        Sets current checked button's ID
        :param value: int
        """

        btn = self._button_group.button(value)
        btn.setChecked(True)
        self.checkedChanged.emit(value)

    theme_checked = Property(int, _get_checked, _set_checked, notify=checkedChanged)


class MenuLineTabWidget(base.BaseWidget, object):
    def __init__(self, alignment=Qt.AlignCenter, parent=None):
        self._alignment = alignment
        super(MenuLineTabWidget, self).__init__(parent=parent)

    @property
    def tool_button_group(self):
        return self._tool_button_group

    def get_main_layout(self):
        main_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))

        return main_layout

    def ui(self):
        super(MenuLineTabWidget, self).ui()

        self._tool_button_group = MenuLineButtonGroup()
        self._bar_layout = layouts.HorizontalLayout(margins=(0, 0, 0, 0))
        if self._alignment == Qt.AlignCenter:
            self._bar_layout.addStretch()
            self._bar_layout.addWidget(self._tool_button_group)
            self._bar_layout.addStretch()
        elif self._alignment == Qt.AlignLeft:
            self._bar_layout.addWidget(self._tool_button_group)
            self._bar_layout.addStretch()
        elif self._alignment == Qt.AlignRight:
            self._bar_layout.addStretch()
            self._bar_layout.addWidget(self._tool_button_group)
        self._stack = stack.SlidingOpacityStackedWidget()
        self._tool_button_group.checkedChanged.connect(self._stack.setCurrentIndex)
        self.main_layout.addLayout(self._bar_layout)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addSpacing(5)
        self.main_layout.addWidget(self._stack)

    def add_tab(self, widget, data_dict):
        """
        Adds a new tab with given data
        :param widget: QWidget
        :param data_dict: dict
        """

        self._stack.addWidget(widget)
        self._tool_button_group.add_button(data_dict, self._stack.count() - 1)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for panel widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, QPoint, QTimer, QEasingCurve, QPropertyAnimation, QAbstractAnimation
from Qt.QtWidgets import QScrollArea

from tpDcc import dcc
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, label, buttons, dividers


class SliderPanelPositions(object):

    LEFT = 'left'
    RIGHT = 'right'
    TOP = 'top'
    BOTTOM = 'bottom'


class SliderPanel(base.BaseWidget, object):
    """
    Panel that slides in from the edge of the window
    """

    closed = Signal()
    closeButtonClicked = Signal()

    def __init__(self, title, position=SliderPanelPositions.RIGHT, closable=True, parent=None):

        self._title = title
        self._position = position
        self._closable = closable
        self._is_first_close = True

        super(SliderPanel, self).__init__(parent)

        self.setObjectName('sliderPanel')
        self.setWindowFlags(Qt.Popup)
        self.setAttribute(Qt.WA_StyledBackground)

        self._close_timer = QTimer(self)
        self._close_timer.setInterval(300)
        self._close_timer.setSingleShot(True)
        self._close_timer.timeout.connect(self.close)
        self._close_timer.timeout.connect(self.closed.emit)

        self._pos_anim = QPropertyAnimation(self)
        self._pos_anim.setTargetObject(self)
        self._pos_anim.setEasingCurve(QEasingCurve.OutCubic)
        self._pos_anim.setDuration(300)
        self._pos_anim.setPropertyName(b'pos')

        self._opacity_anim = QPropertyAnimation()
        self._opacity_anim.setTargetObject(self)
        self._opacity_anim.setDuration(300)
        self._opacity_anim.setEasingCurve(QEasingCurve.OutCubic)
        self._opacity_anim.setPropertyName(b'windowOpacity')
        self._opacity_anim.setStartValue(0.0)
        self._opacity_anim.setEndValue(1.0)

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    @property
    def position(self):
        """
        Returns the placement of the panel in parent window
        :return: str
        """

        return self._position

    @position.setter
    def position(self, value):
        """
        Sets the position of the panel in parent window ('top', 'right', 'bottom' or 'left').
        :param value: str
        """

        self._position = value
        if value in [SliderPanelPositions.BOTTOM, SliderPanelPositions.TOP]:
            self.setFixedHeight(200)
        else:
            self.setFixedWidth(200)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def ui(self):
        super(SliderPanel, self).ui()

        self._title_label = label.BaseLabel(parent=self).h4()
        self._title_label.setText(self._title)

        self._close_btn = buttons.BaseToolButton(parent=self).icon_only().image('close', theme='window').small()
        self._close_btn.setVisible(self._closable or False)

        title_layout = layouts.HorizontalLayout()
        title_layout.addWidget(self._title_label)
        title_layout.addStretch()
        title_layout.addWidget(self._close_btn)

        self._button_layout = layouts.HorizontalLayout()
        self._button_layout.addStretch()

        self._scroll_area = QScrollArea()
        self.main_layout.addLayout(title_layout)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addWidget(self._scroll_area)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addLayout(self._button_layout)

    def setup_signals(self):
        self._close_btn.clicked.connect(self.close)
        self._close_btn.clicked.connect(self.closeButtonClicked.emit)

    def show(self):
        self._update_position()
        self._fade_in()

        return super(SliderPanel, self).show()

    def closeEvent(self, event):
        if self._is_first_close:
            self._is_first_close = False
            self._close_timer.stop()
            self._fade_out()
            self.closed.emit()
            event.ignore()
        else:
            event.accept()

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_widget(self, widget):
        """
        Sets the widget that will be contained inside the panel
        :param widget: QWidget
        """

        self._scroll_area.setWidget(widget)

    def add_button(self, button):
        """
        Adds a new button to the bottom part of the panel
        :param button: QPushButton
        """

        self._button_layout.addWidget(button)

    def left(self):
        """
        Sets the panel's placement to left
        :return: SliderPanel
        """

        self.position = SliderPanelPositions.LEFT

        return self

    def right(self):
        """
        Sets the panel's placement to right
        :return: SliderPanel
        """

        self.position = SliderPanelPositions.RIGHT

        return self

    def top(self):
        """
        Sets the panel's placement to top
        :return: SliderPanel
        """

        self.position = SliderPanelPositions.TOP

        return self

    def bottom(self):
        """
        Sets the panel's placement to bottom
        :return: SliderPanel
        """

        self.position = SliderPanelPositions.BOTTOM

        return self

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _fade_in(self):
        """
        Internal function that fades in the panel
        """

        self._pos_anim.start()
        self._opacity_anim.start()

    def _fade_out(self):
        """
        Internal function that fades out the panel
        """

        self._pos_anim.setDirection(QAbstractAnimation.Backward)
        self._pos_anim.start()
        self._opacity_anim.setDirection(QAbstractAnimation.Backward)
        self._opacity_anim.start()

    def _update_position(self):
        """
        Internal function that makes sure that panel is positioned in the proper place
        """

        parent = self.parent()
        parent_parent = parent.parent()
        dcc_win = dcc.get_main_window()
        dcc_window = parent_parent == dcc_win
        if parent_parent and dcc_win:
            dcc_window = dcc_window or parent_parent.objectName() == dcc_win.objectName()
        parent_geo = parent.geometry()
        if self._position == SliderPanelPositions.LEFT:
            pos = parent_geo.topLeft() if dcc_window else parent.mapToGlobal(parent_geo.topLeft())
            target_x = pos.x()
            target_y = pos.y()
            self.setFixedHeight(parent_geo.height())
            self._pos_anim.setStartValue(QPoint(target_x - self.width(), target_y))
            self._pos_anim.setEndValue(QPoint(target_x, target_y))
        if self._position == SliderPanelPositions.RIGHT:
            pos = parent_geo.topRight() if dcc_window else parent.mapToGlobal(parent_geo.topRight())
            self.setFixedHeight(parent_geo.height())
            target_x = pos.x() - self.width()
            target_y = pos.y()
            self._pos_anim.setStartValue(QPoint(target_x + self.width(), target_y))
            self._pos_anim.setEndValue(QPoint(target_x, target_y))
        if self._position == SliderPanelPositions.TOP:
            pos = parent_geo.topLeft() if dcc_window or parent_parent is None else parent.mapToGlobal(
                parent_geo.topLeft())
            self.setFixedWidth(parent_geo.width())
            target_x = pos.x()
            target_y = pos.y()
            self._pos_anim.setStartValue(QPoint(target_x, target_y - self.height()))
            self._pos_anim.setEndValue(QPoint(target_x, target_y))
        if self._position == SliderPanelPositions.BOTTOM:
            pos = parent_geo.bottomLeft() if dcc_window else parent.mapToGlobal(parent_geo.bottomLeft())
            self.setFixedWidth(parent_geo.width())
            target_x = pos.x()
            target_y = pos.y() - self.height()
            self._pos_anim.setStartValue(QPoint(target_x, target_y + self.height()))
            self._pos_anim.setEndValue(QPoint(target_x, target_y))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom Qt stack widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, QPoint, QEasingCurve, QPropertyAnimation, QParallelAnimationGroup
from Qt.QtWidgets import QFrame, QStackedWidget

from tpDcc.managers import resources
from tpDcc.libs.qt.core import consts, qtutils, base, dpi, mixin
from tpDcc.libs.qt.widgets import layouts, buttons, lineedit


class SlidingStackedWidget(QStackedWidget, object):
    """
    QStackedWidget width sliding functionality
    """

    animFinished = Signal(int)

    def __init__(self, parent=None, **kwargs):
        super(SlidingStackedWidget, self).__init__(parent)

        # This variable holds the current widget without taking into account if the animation has finished or not
        self._speed = kwargs.pop('speed', 500)
        self._current_widget = None
        self._animation_type = kwargs.pop('animation_type', QEasingCurve.OutCubic)
        self._wrap = kwargs.pop('wrap', True)
        self._vertical = kwargs.pop('vertical', False)
        self._active_state = False
        self._blocked_page_list = list()
        self._now = 0
        self._next = 1

    @property
    def current_widget(self):
        return self._current_widget

    def set_speed(self, speed):
        """
        Sets the animation speed of the sliding
        :param speed: int
        """

        self._speed = speed

    def set_animation(self, animation_type):
        """
        Set the curve animation type of the sliding
        :param animation_type: QEasingCurve
        """

        self._animation_type = animation_type

    def set_vertical_mode(self, vertical=True):
        """
        Sets whether the sliding animation is done vertically or horizontally
        :param vertical: bool
        """

        self._vertical = vertical

    def set_wrap(self, wrap):
        """
        Sets whether the page index is restarted when we arrive the last one or not
        :param wrap: bool
        """

        self._wrap = wrap

    def slide_in_next(self):
        """
        Slides into the next widget
        """

        now = self.currentIndex()
        if self._wrap or now < self.count() - 1:
            self.slide_in_index(now + 1)

    def slide_in_prev(self):
        """
        Slides into the previous widget
        """

        now = self.currentIndex()
        if self._wrap or now > 0:
            self.slide_in_index(now - 1)

    def slide_in_index(self, next, force=False):
        """
        Slides to the given widget index
        :param next: int, index of the widget to slide
        """

        now = self.currentIndex()
        if (self._active_state or next == now) and not force:
            return

        self._active_state = True
        width, height = self.frameRect().width(), self.frameRect().height()
        next %= self.count()
        if next > now:
            if self._vertical:
                offset_x, offset_y = 0, height
            else:
                offset_x, offset_y = width, 0
        else:
            if self._vertical:
                offset_x, offset_y = 0, -height
            else:
                offset_x, offset_y = -width, 0
        self.widget(next).setGeometry(0, 0, width, height)
        pnow, pnext = self.widget(now).pos(), self.widget(next).pos()
        self._point_now = pnow

        self.widget(next).move(pnext.x() + offset_x, pnext.y() + offset_y)
        self.widget(next).show()
        self.widget(next).raise_()
        self._current_widget = self.widget(next)

        anim_now = QPropertyAnimation(self.widget(now), b'pos')
        anim_now.setDuration(self._speed)
        anim_now.setStartValue(pnow)
        anim_now.setEndValue(QPoint(pnow.x() - offset_x, pnow.y() - offset_y))
        anim_now.setEasingCurve(self._animation_type)

        anim_next = QPropertyAnimation(self.widget(next), b'pos')
        anim_next.setDuration(self._speed)
        anim_next.setStartValue(QPoint(offset_x + pnext.x(), offset_y + pnext.y()))
        anim_next.setEndValue(pnext)
        anim_next.setEasingCurve(self._animation_type)

        self._anim_group = QParallelAnimationGroup()
        self._anim_group.addAnimation(anim_now)
        self._anim_group.addAnimation(anim_next)
        self._anim_group.finished.connect(self._animation_done_slot)
        self._anim_group.start()

        self._next = next
        self._now = now

    def _animation_done_slot(self):
        self.setCurrentIndex(self._next)
        self.widget(self._now).hide()
        self.widget(self._now).move(self._point_now)
        try:
            self.widget(self._now).update()
        except Exception:
            pass
        self._active_state = False
        self.animFinished.emit(self._next)


@mixin.stacked_opacity_animation_mixin
class SlidingOpacityStackedWidget(QStackedWidget, object):
    """
    Custom stack widget that activates opacity animation when current stack index changes
    """

    def __init__(self, parent=None):
        super(SlidingOpacityStackedWidget, self).__init__(parent)


class StackItem(QFrame, object):

    minimized = Signal()
    maximized = Signal()
    toggleExpandRequested = Signal(bool)
    shiftUpPressed = Signal()
    shiftDownPressed = Signal()
    deletePressed = Signal()
    updateRequested = Signal()

    def __init__(self, title, parent, collapsed=False, collapsable=True, icon=None, start_hidden=False,
                 shift_arrows_enabled=True, delete_button_enabled=True, title_editable=True,
                 icon_size=12, title_frame=None, show_item_icon=True):
        super(StackItem, self).__init__(parent)

        self._stack_widget = parent
        self._icon_size = icon_size
        self._title = title
        self._title_editable = title_editable
        self._icon = icon
        self._shift_arrows_enabled = shift_arrows_enabled
        self._delete_button_enabled = delete_button_enabled
        self._color = consts.DARK_BG_COLOR
        self._contents_margins = (0, 0, 0, 0)
        self._contents_spacing = 0
        self._collapsable = collapsable
        self._collapsed = collapsed
        self._title_frame = title_frame
        self._show_item_icon = show_item_icon

        if start_hidden:
            self.hide()

        self.ui()
        self.setup_signals()

        if not collapsable:
            self._collapsed = False
            self.expand()

        if self._collapsed:
            self.collapse()
        else:
            self.expand()

    @property
    def contents_layout(self):
        return self._contents_layout

    @property
    def collapsed(self):
        return self._collapsed

    def ui(self):
        # theme = tpDcc.ToolsMgr().get_tool_theme('tpDcc-tools-hub')
        # border_width = qtutils.dpi_scale_divide(theme.STACK_BORDER_WIDTH)
        margin = qtutils.dpi_scale_divide(1)
        self.main_layout = layouts.VerticalLayout(margins=(margin, margin, margin, margin), spacing=0)
        # self.main_layout = layouts.VerticalLayout(margins=(0, 0, 0, 0), spacing=0)

        self.setLayout(self.main_layout)

        if not self._title_frame:
            self._title_frame = StackTitleFrame(
                title=self._title, icon=self._icon, title_editable=self._title_editable,
                item_icon_size=self._icon_size, collapsed=self._collapsed,
                shift_arrows_enabled=self._shift_arrows_enabled, delete_button_enabled=self._delete_button_enabled)
        if not self._show_item_icon:
            self._title_frame.item_icon_button.hide()

        self._widget_hider = StackHiderWidget(parent=self)
        self._contents_layout = self._widget_hider.main_layout
        self._contents_layout.setContentsMargins(*self._contents_margins)
        self._contents_layout.setSpacing(self._contents_spacing)
        self._widget_hider.setContentsMargins(0, 0, 0, 0)
        self._widget_hider.setHidden(self._collapsed)
        self._widget_hider.setObjectName('stackbody')

        self.main_layout.addWidget(self._title_frame)
        self.main_layout.addWidget(self._widget_hider)

    def setup_signals(self):
        self._title_frame.expand_toggle_button.leftClicked.connect(self._on_toggle_contents)
        self._title_frame.minimized.connect(self.minimized.emit)
        self._title_frame.maximized.connect(self.maximized.emit)
        self._title_frame.toggleExpandRequested.connect(self.toggleExpandRequested.emit)
        self._title_frame.shiftUpPressed.connect(self.shiftUpPressed.emit)
        self._title_frame.shiftDownPressed.connect(self.shiftDownPressed.emit)
        self._title_frame.deletePressed.connect(self.deletePressed.emit)
        self._title_frame.updateRequested.connect(self.updateRequested.emit)

    def title_text_widget(self):
        """
        Returns title text widget
        :return: QLineEdit
        """

        return self._title_frame.line_edit

    def get_title(self):
        """
        Returns title text
        :return: str
        """

        return self._title_frame.line_edit.text()

    def set_title(self, text):
        """
        Function that sets title text
        :param text: str
        """

        self._title_frame.line_edit.setText(text)

    def add_widget(self, widget):
        """
        Adds a new widget to the contents layout
        :param widget: QWidget
        """

        self._contents_layout.addWidget(widget)

    def add_layout(self, layout):
        """
        Adds a new layout to the contents layout
        :param layout: QLayout
        """

        self._contents_layout.addLayout(layout)

    def expand(self, emit=True):
        """
        Function that expands the contents and show all the widget data
        :param emit: bool
        """

        self._on_expand(emit)

    def collapse(self, emit=True):
        """
        Function that collapses and hides the contents
        :param emit: bool
        """

        self._on_collapse(emit)

    def show_expand_indicator(self, flag):
        """
        Sets whether expand indicator is visible or not
        :param flag: bool
        """

        self._title_frame.expand_toggle_button.setVisible(flag)

    def set_title_text_mouse_transparent(self, flag):
        """
        Sets whether or not title text mouse is transparent
        :param flag: bool
        """

        self._title_frame.line_edit.setAttribute(Qt.WA_TransparentForMouseEvents, flag)

    def set_item_icon_color(self, color):
        """
        Sets the color of the item in title
        :param color: tuple(int, int, int), RGB color in 0-255 range
        """

        self._title_frame.set_item_icon_color(color)

    def set_item_icon(self, icon):
        """
        Sets the icon of the item in title
        :param icon: QIcon
        """

        self.title_frame.set_item_icon(icon)

    def shift_up(self):
        """
        Emits item shift up signal
        """

        self.shiftUpPressed.emit()

    def shift_down(self):
        """
        Emits shijft down signal
        """

        self.shiftDownPressed.emit()

    def set_arrows_visible(self, flag):
        """
        Sets whether shift arrows are visible or not.
        This arrows allow the item to be shifted upwards or downwards the stack
        :param flag: bool
        """

        if flag:
            self._title_frame.shift_down_button.show()
            self._title_frame.shift_up_button.show()
        else:
            self._title_frame.shift_down_button.hide()
            self._title_frame.shift_up_button.hide()

    def update_size(self):
        """
        Updates the size of the widget, to fit to new size depending on its contents
        """

        self.updateRequested.emit()

    def _on_expand(self, emit=True):
        """
        Internal function that expands the contents and show all the widget data
        :param emit: bool
        """

        self._widget_hider.setHidden(False)
        self._title_frame.expand()
        if emit:
            self.maximized.emit()
        self._collapsed = False

    def _on_collapse(self, emit=True):
        """
        Internal function that collapses and hides the contents
        :param emit: bool
        """

        self._widget_hider.setHidden(True)
        self._title_frame.collapse()
        if emit:
            self.minimized.emit()
        self._collapsed = True

    def _on_toggle_contents(self, emit=True):
        """
        Internal callback function that shows and hides the hidder widget and shows/hides contents
        :param emit: bool
        """

        if not self._collapsable:
            return

        self.toggleExpandRequested.emit(not self._collapsed)
        if self._collapsed:
            self.expand(emit)
            self.update_size()
            return not self._collapsed

        self.collapse(emit)
        self.update_size()
        return self._collapsed


class StackHiderWidget(base.BaseFrame, object):
    def __init__(self, parent=None):
        super(StackHiderWidget, self).__init__(parent)


class StackTitleFrame(QFrame, dpi.DPIScaling):

    minimized = Signal()
    maximized = Signal()
    toggleExpandRequested = Signal(bool)
    shiftUpPressed = Signal()
    shiftDownPressed = Signal()
    deletePressed = Signal()
    updateRequested = Signal()

    def __init__(self, title='', title_editable=False, icon=None, item_icon_size=16, collapsed=True,
                 shift_arrows_enabled=True, delete_button_enabled=True, parent=None):
        super(StackTitleFrame, self).__init__(parent)

        self._title = title
        self._item_icon = icon or 'tpdcc'
        self._title_editable = title_editable
        self._shift_arrows_enabled = shift_arrows_enabled
        self._delete_button_enabled = delete_button_enabled
        self._spaces_to_underscore = False
        self._collapsed = collapsed
        self._item_icon_size = item_icon_size

        self._collapsed_icon = resources.icon('sort_closed')
        self._expand_icon = resources.icon('sort_down')

        self.setObjectName('title')

        self.ui()
        self.setup_signals()

        self.setFixedHeight(self.sizeHint().height())
        self.setMinimumSize(self.sizeHint().width(), self.sizeHint().height() + 1)

    @property
    def horizontal_layout(self):
        return self._horizontal_layout

    @property
    def item_icon(self):
        return self._item_icon

    @property
    def item_icon_button(self):
        return self._item_icon_btn

    @property
    def expand_toggle_button(self):
        return self._expand_toggle_btn

    @property
    def shift_down_button(self):
        return self._shift_down_btn

    @property
    def shift_up_button(self):
        return self._shift_up_btn

    @property
    def delete_button(self):
        return self._delete_btn

    @property
    def line_edit(self):
        return self._line_edit

    def mouseDoubleClickEvent(self, event):
        if self._title_editable:
            self._line_edit.editEvent(event)

    def ui(self):

        item_icon = resources.icon(self._item_icon)
        if item_icon.isNull():
            item_icon = resources.icon('tpdcc')
        delete_icon = resources.icon('delete')
        shift_up_icon = resources.icon('arrow_up')
        shift_down_icon = resources.icon('arrow_down')

        self.setContentsMargins(*qtutils.margins_dpi_scale(0, 0, 0, 0))
        self._extras_layout = layouts.HorizontalLayout(margins=(0, 0, 0, 0), spacing=0)
        self._horizontal_layout = layouts.HorizontalLayout(spacing=0)
        self.setLayout(self._horizontal_layout)

        self._line_edit = lineedit.ClickLineEdit(self._title)
        self._line_edit.setObjectName('lineEdit')
        self._line_edit.setFocusPolicy(Qt.NoFocus)
        self._line_edit.setVisible(False)
        self._line_edit.setAttribute(Qt.WA_TransparentForMouseEvents)
        if not self._title_editable:
            self._line_edit.setReadOnly(True)

        self._expand_toggle_btn = buttons.BaseMenuButton()
        self._item_icon_btn = buttons.BaseMenuButton()
        self._shift_down_btn = buttons.BaseMenuButton()
        self._shift_up_btn = buttons.BaseMenuButton()
        self._delete_btn = buttons.BaseMenuButton()

        if not self._shift_arrows_enabled:
            self._shift_down_btn.hide()
            self._shift_up_btn.hide()
        if not self._delete_button_enabled:
            self._delete_btn.hide()

        self._item_icon_btn.set_icon(icon=item_icon, size=self._item_icon_size)
        self._item_icon_btn.setAttribute(Qt.WA_TransparentForMouseEvents)

        icon_size = 12
        highlight_offset = 40
        self._delete_btn.set_icon(delete_icon, colors=None, size=icon_size, color_offset=highlight_offset)
        self._shift_up_btn.set_icon(shift_up_icon, colors=None, size=icon_size, color_offset=highlight_offset)
        self._shift_down_btn.set_icon(shift_down_icon, colors=None, size=icon_size, color_offset=highlight_offset)
        self._expand_toggle_btn.set_icon(self._expand_icon, colors=(192, 192, 192), size=icon_size)

        self._horizontal_layout.addWidget(self._item_icon_btn)
        self._horizontal_layout.addStretch()
        self._horizontal_layout.addWidget(self._line_edit)
        self._horizontal_layout.addLayout(self._extras_layout)
        self._horizontal_layout.addWidget(self._shift_up_btn)
        self._horizontal_layout.addWidget(self._shift_down_btn)
        self._horizontal_layout.addWidget(self._expand_toggle_btn)
        self._horizontal_layout.addWidget(self._delete_btn)
        self._horizontal_layout.setStretchFactor(self._line_edit, 4)

    def setup_signals(self):
        self._shift_up_btn.leftClicked.connect(self._on_shift_up)
        self._shift_down_btn.leftClicked.connect(self._on_shift_down)
        self._delete_btn.leftClicked.connect(self.deletePressed.emit)
        self._line_edit.textChanged.connect(self._on_title_validate)
        self._line_edit.selectionChanged.connect(self._on_select_check)

    def collapse(self):
        self._expand_toggle_btn.set_icon(self._collapsed_icon, colors=(192, 192, 192), size=12)

    def expand(self):
        self._expand_toggle_btn.set_icon(self._expand_icon, colors=(192, 192, 192), size=12)

    def set_item_icon_color(self, color):
        self._item_icon_btn.set_icon_color(color)

    def set_item_icon(self, icon):
        self._item_icon_btn.set_icon(icon)

    def _on_shift_up(self):
        self.shiftUpPressed.emit()

    def _on_shift_down(self):
        self.shiftDownPressed.emit()

    def _on_title_validate(self):
        """
        Internal callback function that is called when line edit text changes
        Removes invalid characters and replaces spaces with underscores
        """

        if self._spaces_to_underscore:
            line_edit = self._line_edit
            text = self._line_edit.text()
            pos = line_edit.cursorPosition()
            text = text.replace(' ', '_')
            line_edit.blockSignals(True)
            try:
                line_edit.setText(text)
            finally:
                line_edit.blockSignals(False)
            line_edit.setCursorPosition(pos)

    def _on_select_check(self):
        """
        Internal callback function that is called when user deselects line edit
        Stops any selection from happening if title is not editable
        """

        if not self._title_editable:
            self._line_edit.deselect()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains form widgets for library
"""

from __future__ import print_function, division, absolute_import

import logging
from functools import partial

from Qt.QtCore import Signal
from Qt.QtWidgets import QSizePolicy, QFrame, QSpacerItem

from tpDcc.libs.qt.core import contexts as qt_contexts
from tpDcc.libs.qt.widgets import layouts, label, buttons, formfields

from tpDcc.tools.datalibrary.core import settings

LOGGER = logging.getLogger('tpDcc-libs-qt')


class FormDialog(QFrame, object):

    accepted = Signal(object)
    rejected = Signal(object)

    def __init__(self, parent=None, form=None):
        super(FormDialog, self).__init__(parent)

        self._settings = None

        self.main_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self.setLayout(self.main_layout)

        self._widgets = list()
        self._validator = None

        self._title = label.BaseLabel(parent=self)
        self._title.setObjectName('title')
        self._title.setText('FORM')
        self._description = label.BaseLabel(parent=self)
        self._description.setObjectName('description')
        self._form_widget = FormWidget(self)
        self._form_widget.setObjectName('formWidget')
        self._form_widget.validated.connect(self._on_validated)
        btn_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._accept_btn = buttons.BaseButton(parent=self)
        self._accept_btn.setObjectName('acceptButton')
        self._accept_btn.setText('Accept')
        self._accept_btn.clicked.connect(self.accept)
        self._reject_btn = buttons.BaseButton(parent=self)
        self._reject_btn.setObjectName('rejectButton')
        self._reject_btn.setText('Cancel')
        self._reject_btn.clicked.connect(self.reject)
        btn_layout.addSpacerItem(QSpacerItem(10, 0, QSizePolicy.Expanding, QSizePolicy.Preferred))
        btn_layout.addWidget(self._accept_btn)
        btn_layout.addWidget(self._reject_btn)

        self.main_layout.addWidget(self._title)
        self.main_layout.addWidget(self._description)
        self.main_layout.addWidget(self._form_widget)
        self.main_layout.addStretch(1)
        self.main_layout.addLayout(btn_layout)

        if form:
            self.set_settings(form)

    def accept_button(self):
        """
        Returns the accept button
        :return: QPushButton
        """

        return self._accept_btn

    def reject_button(self):
        """
        Returns the reject button
        :return: QPushButton
        """

        return self._reject_btn

    def set_settings(self, settings):
        """
        Sets dialog form settings
        :param settings: dict
        """

        self._settings = settings
        title = settings.get("title")
        if title is not None:
            self._title.setText(title)
        callback = settings.get("accepted")
        if not callback:
            self._settings["accepted"] = self._validate_accepted
        callback = settings.get("rejected")
        if not callback:
            self._settings["rejected"] = self._validate_rejected
        description = settings.get("description")
        if description is not None:
            self._description.setText(description)
        validator = settings.get("validator")
        if validator is not None:
            self._form_widget.set_validator(validator)
        layout = settings.get("layout")
        schema = settings.get("schema")
        if schema is not None:
            self._form_widget.set_schema(schema, layout=layout)

    def accept(self):
        """
        Function called when the dialog is accepted
        """

        callback = self._settings.get('accepted')
        if callback:
            callback(**self._form_widget.values())
        self.close()

    def reject(self):
        """
        Function called when the dialog is rejected
        """

        callback = self._settings.get('rejected')
        if callback:
            callback(**self._form_widget.default_values())
        self.close()

    def _validate_accepted(self, **kwargs):
        """
        Internal function called when the accept button has been clicked
        :param kwargs: dict, default values of the form fields
        """

        self._form_widget.validator()(**kwargs)

    def _validate_rejected(self, **kwargs):
        """
        Internal function called when reject button has been clicked
        :param kwargs: dict, default values of the form fields
        """

        self._form_widget.validator()(**kwargs)

    def _on_validated(self):
        """
        Internal callback function that is triggered when the has been validated
        """

        self._accept_btn.setEnabled(not self._form_widget.has_errors())


class FormWidget(QFrame, object):

    accepted = Signal(object)
    stateChanged = Signal()
    validated = Signal()

    def __init__(self, *args, **kwargs):
        super(FormWidget, self).__init__(*args, **kwargs)

        self._schema = dict()
        self._widgets = list()
        self._validator = None

        main_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self.setLayout(main_layout)

        self._fields_frame = QFrame(self)
        self._fields_frame.setObjectName('fieldsFrame')
        options_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._fields_frame.setLayout(options_layout)

        self._title_widget = buttons.BaseButton(parent=self)
        self._title_widget.setCheckable(True)
        self._title_widget.setObjectName('titleWidget')
        self._title_widget.toggled.connect(self._on_title_clicked)
        self._title_widget.hide()

        main_layout.addWidget(self._title_widget)
        main_layout.addWidget(self._fields_frame)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def closeEvent(self, event):
        self.save_persistent_values()
        super(FormWidget, self).closeEvent(event)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def title_widget(self):
        """
        Returns the title widget
        :return: QWidget
        """

        return self._title_widget

    def set_title(self, title):
        """
        Sets the title text
        :param title: str
        """

        self.title_widget().setText(title)

    def is_expanded(self):
        """
        Returns whether the item is expanded or not
        :return: bool
        """

        return self._title_widget.isChecked()

    def set_expanded(self, flag):
        """
        Expands the options if True, otherwise collapses the options
        :param flag: bool
        """

        with qt_contexts.block_signals(self._title_widget):
            self._title_widget.setChecked(flag)
            self._fields_frame.setVisible(flag)

    def set_title_visible(self, flag):
        """
        Sets whether the title widget is visible or not
        :param flag: bool
        """

        self.title_widget().setVisible(flag)

    def widget(self, name):
        """
        Returns the widget for the given widget name
        :param name: str
        :return: FieldWidget
        """

        for widget in self._widgets:
            if widget.data().get('name') == name:
                return widget

    def value(self, name):
        """
        Returns the value for the given widget name
        :param name: str
        :return: object
        """

        widget = self.widget(name)
        if not widget:
            return None

        return widget.value()

    def set_value(self, name, value):
        """
        Sets the value for the given field name
        :param name: str
        :param value: variant
        """

        widget = self.widget(name)
        widget.set_value(value)

    def values(self):
        """
        Returns all the field values indexed by the field name
        :return: dict
        """

        values = dict()
        for widget in self._widgets:
            name = widget.data().get('name')
            if name:
                values[name] = widget.value()

        return values

    def set_values(self, values):
        """
        Sets the field values for the current form
        :param values: dict
        """

        state = list()
        for name in values:
            state.append({'name': name, 'value': values[name]})

        self._set_state(state)

    def default_values(self):
        """
        Returns all teh default field values indexed by the field name
        :return: dict
        """

        values = dict()
        for widget in self._widgets:
            name = widget.data().get('name')
            if name:
                values[name] = widget.default()

        return values

    def set_data(self, name, data):
        """
        Sets the data for the given field name
        :param name: str
        :param data: dict
        """

        widget = self.widget(name)
        if not widget:
            return
        widget.set_data(data)

    def fields(self):
        """
        Returns fields data for the form
        :return: list(dict)
        """

        options = list()
        for widget in self._widgets:
            options.append(widget.data())

        return options

    def field_widgets(self):
        """
        Returns all field widgets
        :return: list(FieleWidget)
        """

        return self._widgets

    def state(self):
        """
        Returns the current state
        :return: dict
        """

        fields = list()
        for widget in self._widgets:
            fields.append(widget.state())

        state = {
            'fields': fields,
            'expanded': self.is_expanded()
        }

        return state

    def set_state(self, state):
        """
        Sets the current state
        :param state: dict
        """

        expanded = state.get('expanded')
        if expanded is not None:
            self.set_expanded(expanded)

        fields = state.get('fields')
        if fields is not None:
            self._set_state(fields)

        self.validate()

    def schema(self):
        """
        Returns form's schema
        :return: dict
        """

        return self._schema

    def set_schema(self, schema, layout=None, errors_visible=False):
        """
        Sets the schema for the widget
        :param schema: list(dict)
        :param layout: str
        :param errors_visible: str
        """

        self._schema = self._sort_schema(schema)
        if not self._schema:
            return

        for field in schema:
            cls = formfields.FIELD_WIDGET_REGISTRY.get(field.get('type', 'label'))
            if not cls:
                LOGGER.warning('Cannot find widget for {}'.format(field))
                continue
            if layout and not field.get('layout'):
                field['layout'] = layout

            enabled = field.get('enabled', True)
            read_only = field.get('readOnly', False)

            error_visible = field.get('errorVisible')
            field['errorVisible'] = error_visible if error_visible is not None else errors_visible

            widget = cls(data=field, parent=self._fields_frame, form_widget=self)
            data = widget.default_data()
            data.update(field)

            widget.set_data(data)

            value = field.get('value')
            default = field.get('default')
            if value is None and default is not None:
                widget.set_value(default)

            if not enabled or read_only:
                widget.setEnabled(False)

            self._widgets.append(widget)

            callback = partial(self._on_field_changed, widget)
            widget.valueChanged.connect(callback)

            self._fields_frame.layout().addWidget(widget)

        self.load_persistent_values()

    def validator(self):
        """
        Returns the validator for the form
        :return: fn
        """

        return self._validator

    def set_validator(self, validator):
        """
        Sets the validator for the options
        :param validator: fn
        """

        self._validator = validator

    def reset(self):
        """
        Reset all option widget back to the ir default values
        """

        for widget in self._widgets:
            widget.reset()
        self.validate()

    def validate(self, widget=None):
        """
        Validates the current options using the validator
        """

        if not self._validator:
            return

        values = dict()
        for name, value in self.values().items():
            data = self.widget(name).data()
            if data.get('validate', True):
                values[name] = value

        if widget:
            values['fieldChanged'] = widget.name()

        fields = self._validator(**values)
        if fields is not None:
            self._set_state(fields)

        self.validated.emit()

    def errors(self):
        """
        Returns all form errors
        :return: list(str)
        """

        errors = list()
        for widget in self._widgets:
            error = widget.data().get('error')
            if error:
                errors.append(error)

        return errors

    def has_errors(self):
        """
        Returns whether the form contains any error
        :return: bool
        """

        return bool(self.errors())

    def save_persistent_values(self):
        """
        Saves form widget values
        Triggered when the user changes field values
        """

        data = dict()

        for widget in self._widgets:
            name = widget.data().get('name')
            if name and widget.data().get('persistent'):
                key = self.objectName() or 'FormWidget'
                key = widget.data().get('persistentKey', key)
                data.setdefault(key, dict())
                data[key][name] = widget.value()

        for key in data:
            settings.set(key, data[key])

    def load_persistent_values(self):
        """
        Returns the options from the user settings
        :return: dict
        """

        values = dict()
        default_values = self.default_values()

        for field in self.schema():
            name = field.get('name')
            persistent = field.get('persistent')
            if persistent:
                key = self.objectName() or 'FormWidget'
                key = field.get('persistentKey', key)
                value = settings.get(key, dict()).get(name)
            else:
                value = default_values.get(name)

            if value is not None:
                values[name] = value

        self.set_values(values)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _sort_schema(self, schema):
        """
        Internal function that sorts the schema depending on the group order
        :param schema: list(dict)
        :return: list(dict)
        """

        def _key(field):
            return field['order']

        order = 0

        if not schema:
            return

        for i, field in enumerate(schema):
            if field.get('type') == 'group':
                order = field.get('order', order)
            field['order'] = order

        return sorted(schema, key=_key)

    def _set_state(self, fields):
        """
        Internal function that sets fields state
        :param fields: list(dict)
        """

        for widget in self._widgets:
            widget.blockSignals(True)

        try:
            for widget in self._widgets:
                widget.set_error('')
                for field in fields:
                    if field.get('name') == widget.data().get('name'):
                        widget.set_data(field)
        finally:
            for widget in self._widgets:
                widget.blockSignals(False)

        self.stateChanged.emit()

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_title_clicked(self, toggle):
        """
        Internal callback function that is triggered when the user clicks in the title widget
        """

        self.set_expanded(toggle)
        self.stateChanged.emit()

    def _on_field_changed(self, widget):
        """
        Internal callback function triggered when the given option widget changes its value
        :param widget: FieldWidget
        """

        self.validate(widget=widget)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different graphic views
"""

from __future__ import print_function, division, absolute_import

import os
import math
import logging

from Qt.QtCore import Qt, Signal, QPoint, QRectF, QLineF
from Qt.QtWidgets import QGraphicsRectItem, QGraphicsView, QGraphicsItem
from Qt.QtGui import QColor, QPen, QBrush, QPainter, QImage, QVector2D

from tpDcc.libs.python import mathlib

LOGGER = logging.getLogger('tpDcc-libs-qt')

try:
    from Qt import QtOpenGL
except ImportError:
    try:
        from PySide import QtOpenGL
    except ImportError:
        # Max 2018 does not support QtOpenGL?
        LOGGER.warning('QtOpenGL is not available! QtOpenGL functionality will not be available!')


class ViewportModes(object):
    Full = 'full'
    Smart = 'smart'
    Minimal = 'minimal'
    Bounding = 'bounding'


class AutoPanController(object):
    def __init__(self, amount=10.0):
        super(AutoPanController, self).__init__()

        self._enabled = False
        self._amount = amount
        self._auto_pan_delta = QVector2D(0.0, 0.0)
        self._been_outside = False

    def get_amount(self):
        return self._amount

    def set_amount(self, amount):
        self._amount = amount

    def get_auto_pan_delta(self):
        return self._auto_pan_delta

    def get_enabled(self):
        return self._enabled

    amount = property(get_amount, set_amount)
    auto_pan_delta = property(get_auto_pan_delta)
    enabled = property(get_enabled)

    def start(self):
        self._enabled = True

    def update(self, rect, pos):
        if self._enabled:
            if pos.x() < 0:
                self._auto_pan_delta = QVector2D(-self._amount, 0.0)
                self._been_outside = True
                self._amount = mathlib.clamp(abs(pos.x()) * 0.3, 0.0, 25.0)
            if pos.x() > rect.width():
                self._auto_pan_delta = QVector2D(self._amount, 0.0)
                self._been_outside = True
                self._amount = mathlib.clamp(abs(rect.width() - pos.x()) * 0.3, 0.0, 25.0)
            if pos.y() < 0:
                self._auto_pan_delta = QVector2D(0.0, -self._amount)
                self._been_outside = True
                self._amount = mathlib.clamp(abs(pos.y()) * 0.3, 0.0, 25.0)
            if pos.y() > rect.height():
                self._auto_pan_delta = QVector2D(0.0, self._amount)
                self._been_outside = True
                self._amount = mathlib.clamp(abs(rect.height() - pos.y()) * 0.3, 0.0, 25.0)
            if self._been_outside and rect.contains(pos):
                self.reset()

    def stop(self):
        self._enabled = False
        self.reset()

    def reset(self):
        self._been_outside = False
        self._auto_pan_delta = QVector2D(0.0, 0.0)


class RubberRect(QGraphicsRectItem, object):

    DEFAULT_RUBBER_RECT_COLOR = QColor(255, 255, 255, 50)

    def __init__(self, name):
        super(RubberRect, self).__init__()
        self._name = name
        self.setZValue(2)
        self.setPen(QPen(self.DEFAULT_RUBBER_RECT_COLOR, 0.5, Qt.SolidLine))
        self.setBrush(QBrush(self.DEFAULT_RUBBER_RECT_COLOR))

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    name = property(get_name, set_name)


class BaseGraphicsView(QGraphicsView, object):
    """
    QGraphicsView with custom functionality:
        - Zoom in/Zoom out
        - Panning/Auto Panning
        - Custom Rubber Rect Selection
    """

    selectionChanged = Signal()

    def __init__(self, min_scale=0.5, max_scale=2.0, parent=None, **kwargs):
        super(BaseGraphicsView, self).__init__(parent=parent)

        self._min_scale = min_scale
        self._max_scale = max_scale
        self._factor = 1.0
        self._factor_diff = 0.0
        self._scale = 1.0
        self._pan_speed = 1.0
        self._is_panning = False
        self._init_scrollbars_pos = QVector2D(self.horizontalScrollBar().value(), self.verticalScrollBar().value())
        self._is_rubber_rect_selection = False
        self._use_opengl = kwargs.get('use_opengl', False)
        self.log = kwargs.get('log', LOGGER)

        self._mouse_pressed = False
        self._pressed_item = None
        self._released_item = None
        self._mouse_pressed_pos = QPoint(0, 0)
        self._mouse_release_pos = QPoint(0, 0)
        self._mouse_pos = QPoint(0, 0)
        self._last_mouse_pos = QPoint(0, 0)

        self._auto_pan_controller = AutoPanController()
        self._rubber_rect = RubberRect(name='RubberRect')

        self.initialize_scene_view()

        self.setInteractive(True)
        self.setMouseTracking(True)
        self.setRenderHint(QPainter.Antialiasing, True)
        self.setRenderHint(QPainter.TextAntialiasing, True)
        self.setRenderHint(QPainter.HighQualityAntialiasing, True)
        self.setRenderHint(QPainter.SmoothPixmapTransform, True)
        self.setRenderHint(QPainter.NonCosmeticDefaultPen, True)
        self.setAttribute(Qt.WA_AlwaysShowToolTips)
        self.setRubberBandSelectionMode(Qt.IntersectsItemShape)

    # region Properties
    def get_viewport_mode(self):
        mode = self.viewportUpdateMode()
        if mode == QGraphicsView.FullViewportUpdate:
            return ViewportModes.Full
        if mode == QGraphicsView.SmartViewportUpdate:
            return ViewportModes.Smart
        if mode == QGraphicsView.MinimalViewport:
            return ViewportModes.Minimal
        if mode == QGraphicsView.BoundingRectViewportUpdate:
            return ViewportModes.Bounding

        return None

    def set_viewport_mode(self, mode):
        if mode == ViewportModes.Full:
            mode = QGraphicsView.FullViewportUpdate
        if mode == ViewportModes.Smart:
            mode = QGraphicsView.SmartViewportUpdate
        if mode == ViewportModes.Minimal:
            mode = QGraphicsView.MinimalViewportUpdate
        if mode == ViewportModes.Bounding:
            mode = QGraphicsView.BoundingRectViewportUpdate

        self.setViewportUpdateMode(mode)
    # endregion

    # region Override Functions
    def mousePressEvent(self, event):
        super(BaseGraphicsView, self).mousePressEvent(event)

        self._pressed_item = self.itemAt(event.pos())
        self._mouse_pressed_pos = event.pos()

        modifiers = event.modifiers()

        if self._pressed_item and isinstance(self._pressed_item, QGraphicsItem):
            self._auto_pan_controller.start()

        if not self._pressed_item:
            if event.button() == Qt.LeftButton:
                self._is_rubber_rect_selection = True
            if event.button() == Qt.RightButton and modifiers == Qt.NoModifier:
                self._is_panning = True
            self._init_scrollbars_pos = QVector2D(self.horizontalScrollBar().value(), self.verticalScrollBar().value())

    def mouseMoveEvent(self, event):
        super(BaseGraphicsView, self).mouseMoveEvent(event)

        self._mouse_pos = event.pos()

        if self._is_panning:
            delta = self.mapToScene(event.pos()) - self.mapToScene(self._last_mouse_pos)
            self._pan(delta)

        if self._rubber_rect and self.scene():
            if self._is_rubber_rect_selection:
                current_pos = self.mapToScene(self._mouse_pos)
                press_pos = self.mapToScene(self._mouse_pressed_pos)
                if self._rubber_rect not in self.scene().items():
                    self.scene().addItem(self._rubber_rect)
                if not self._rubber_rect.isVisible():
                    self._rubber_rect.setVisible(True)
                r = QRectF(
                    press_pos.x(), press_pos.y(), current_pos.x() - press_pos.x(), current_pos.y() - press_pos.y())
                self._rubber_rect.setRect(r.normalized())

        self._auto_pan_controller.update(self.viewport().rect(), event.pos())

        self._last_mouse_pos = event.pos()

    def mouseReleaseEvent(self, event):
        super(BaseGraphicsView, self).mouseReleaseEvent(event)

        self._auto_pan_controller.stop()
        self._mouse_release_pos = event.pos()
        self._released_item = self.itemAt(event.pos())
        self._is_panning = False

        if self._rubber_rect and self.scene():
            if self._is_rubber_rect_selection:
                self._is_rubber_rect_selection = False
                self.setDragMode(QGraphicsView.NoDrag)
                self._select_rubber_rect_items()
                self.remove_item_by_name(self._rubber_rect.name)

    def keyPressEvent(self, event):
        modifiers = event.modifiers()

        # Zoom In
        if all([event.key() == Qt.Key_Equal, modifiers == Qt.ControlModifier]):
            # TODO: Check why is not working
            self.zoom_delta(True)

        # Zoom Out
        if all([event.key() == Qt.Key_Minus, modifiers == Qt.ControlModifier]):
            self.zoom_delta(False)

        # Rest Graph Zoom
        if all([event.key() == Qt.Key_R, modifiers == Qt.ControlModifier]):
            self.reset_scale()

        super(BaseGraphicsView, self).keyPressEvent(event)

    def wheelEvent(self, event):
        self._zoom(math.pow(2.0, event.delta() / 240.0))
    # endregion

    # region Public Functions
    def initialize_scene_view(self):
        """
        Instantiates the proper Scene and enables OpenGL if necessary
        Scene instantiation must be override in child classes
        """

        if self._use_opengl:
            self.setViewport(QtOpenGL.QGLWidget(QtOpenGL.QGLFormat(QtOpenGL.QGL.SampleBuffers)))
            self.log.info('Initializing OpenGL renderer')

    def get_contents_size(self):
        """
        Returns the view contents size (pyshical size)
        :return: tuple, content size
        """

        content_rect = self.contentsRect()
        return (content_rect.width(), content_rect.height())

    def get_center(self):
        """
        Returns the correct center point of the current view
        :return: QPointF, current view center point
        """

        return self.mapToScene(self.viewport().rect().center())

    def set_center(self, pos):
        """
        Sets the current scene center point
        :param pos: tuple, x and y coordinates
        """

        self.centerOn(pos[0], pos[1])

    def get_scene_coordinates(self):
        """
        Returns the scene size
        :return: tuple, coordinates of current scene (-x, -y, x, y)
        """

        if self.scene():
            return self.scene().sceneRect().getCoords()
        return (0, 0, 0, 0)

    def get_translation(self):
        """
        Returns the current scrollbar positions
        :return: tuple, scroll bar coordinates (h, v)
        """

        return [self.horizontalScrollBar().value(), self.verticalScrollBar().value()]

    def get_scale_factor(self):
        """
        Returns the current scale factor
        """

        return [self.transform().m11(), self.transform().m22()]

    def get_scene_attributes(self):
        """
        Returns a dictionary of scene attributes
        :return: dict, view position, scale, size
        """

        scene_attributes = dict()
        scene_attributes.update(view_scale=self.get_scale_factor())
        scene_attributes.update(view_center=self.get_center())
        scene_attributes.update(view_size=self.get_contents_size())
        scene_attributes.update(scene_size=self.get_scene_coordinates())

        return scene_attributes

    def update_scene_attributes(self):
        """
        Updates scene attributes
        """

        scene_attributes = self.get_scene_attributes()
        return scene_attributes

    def remove_item_by_name(self, name):
        """
        Remove graph item by its name (can be any kind of QGraphicsItem)
        :param name: str, name of the item to delete
        """

        if self.scene():
            [self.scene().removeItem(i) for i in self.scene().items() if hasattr(i, 'name') and i.name == name]

    def move_scrollbar(self, delta):
        """
        Move the scrollbar of the view by a given delta
        :param delta: QPointF
        """

        x = self.horizontalScrollBar().value() + delta.x()
        y = self.verticalScrollBar().value() + delta.y()
        self.horizontalScrollBar().setValue(x)
        self.verticalScrollBar().setValue(y)

    def set_scrollbars_positions(self, horizontal, vertical):
        """
        Set the position of the view scrollbars
        :param horizontal: float, position of the horizontal scrollbar
        :param vertical: float, position of the vertical scrollbar
        :return:
        """

        try:
            self.horizontalScrollBar().setValue(horizontal)
            self.verticalScrollBar().setValue(vertical)
        except Exception as e:
            LOGGER.debug(str(e))

    def reset_scale(self):
        self.resetMatrix()

    def zoom_delta(self, direction):
        if direction:
            self._zoom(1 + 0.1)
        else:
            self._zoom(1 - 0.1)

    def fit_scene_content(self):
        """
        Fits scene content to view, by scaling it
        """

        if self.scene():
            scene_rect = self.scene().get_bounding_rect(margin=0)
            self.fitInView(scene_rect, Qt.KeepAspectRatio)
    # endregion

    # region Private Functions
    def _pan(self, delta):
        delta *= self._scale * -1
        delta *= self._pan_speed
        self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() + delta.x())
        self.verticalScrollBar().setValue(self.verticalScrollBar().value() + delta.y())

    def _zoom(self, scale_factor):
        self._factor = round(self.matrix().scale(scale_factor, scale_factor).mapRect(QRectF(0, 0, 1, 1)).width(), 1)

        if hasattr(self.scene(), 'grid_size'):
            if self._factor < (self._min_scale + 0.4):
                self.scene().grid_size = 20
            else:
                self.scene().grid_size = 10

        if self._factor < self._min_scale or self._factor > self._max_scale:
            return
        self.scale(scale_factor, scale_factor)
        self._scale *= scale_factor

    def _select_rubber_rect_items(self):
        self.scene().blockSignals(True)
        items = [i for i in self._rubber_rect.collidingItems()]
        for item in items[:-1]:
            item.setSelected(True)
        self.scene().blockSignals(False)
        if len(items) > 0:
            items[-1].setSelected(True)


class GridView(BaseGraphicsView, object):
    """
    View with grid drawing support
    """

    def __init__(self, parent=None):
        super(GridView, self).__init__(parent=parent)

        self._grid_spacing = 100
        self._grid_size = 10
        self._draw_grid_size = self._grid_size * 2
        self._show_grid = True

        self.setRenderHint(QPainter.Antialiasing)

        self.setViewportUpdateMode(QGraphicsView.BoundingRectViewportUpdate)
        self.setCacheMode(QGraphicsView.CacheBackground)
        self.setAttribute(Qt.WA_AlwaysShowToolTips)

    def drawBackground(self, painter, rect):
        super(GridView, self).drawBackground(painter, rect)

        if self._show_grid:
            scene_rect = self.sceneRect()
            left = int(scene_rect.left()) - (int(scene_rect.left()) % self._draw_grid_size)
            top = int(scene_rect.top()) - (int(scene_rect.top()) % self._draw_grid_size)
            scale_mult = 1.0

            lines = list()
            left_line = rect.left() - rect.left() % self._grid_size
            top_line = rect.top() - rect.top() % self._grid_size
            i = int(left_line)
            while i < int(rect.right()):
                lines.append(QLineF(i, rect.top(), i, rect.bottom()))
                i += self._grid_size
            u = int(top_line)
            while u < int(rect.bottom()):
                lines.append(QLineF(rect.left(), u, rect.right(), u))
                u += self._grid_size
            # TODO: Change pen to a class variable (avoid to create a pen each drawing frame)
            pen = QPen()
            pen.setWidth(0)
            pen.setColor(QColor(20, 20, 20))
            painter.setPen(pen)
            painter.drawLines(lines)


class GridBackgroundImageView(GridView, object):
    """
    View with image background drawing support
    """

    def __init__(self, parent=None):
        super(GridBackgroundImageView, self).__init__(parent=parent)

        self._background_image = None
        self._background_image_path = None
        self._fit_image_to_window = False

        self.setBackgroundBrush(QBrush(QColor(70, 70, 70, 255)))

    def get_background_image(self):
        return self._background_image

    def set_background_image(self, value, fit_to_window=False, mirror_x=False, mirror_y=False):
        if not value:
            return
        value = str(value)
        if not (value and os.path.exists(value)):
            print('background image not found: {}'.format(value))
            return

        self._background_image_path = value
        self._fit_image_to_window = fit_to_window

        # Load image and mirror it vertically
        self._background_image = QImage(value).mirrored(mirror_x, mirror_y)

        if not fit_to_window:
            width = self._background_image.width()
            height = self._background_image.height()
            self.scene().set_size(width, height)
            self.fit_scene_content()

    background_image = property(get_background_image, set_background_image)

    # def resizeEvent(self, event):
    # TODO: tpoveda: This is not working when an image is not loaded
    #     if self._fit_image_to_window:
    #         self.scene().set_size(self.rect().width(), self.rect().height())
    #     else:
    #         self.fit_scene_content()
    #     super(GridBackgroundImageView, self).resizeEvent(event)

    def drawBackground(self, painter, rect):
        """
        Override method to draw view custom background image
        """

        if not self.background_image:
            return super(GridBackgroundImageView, self).drawBackground(painter, rect)
        super(GridBackgroundImageView, self).drawBackground(painter, rect)
        painter.drawImage(self.sceneRect(), self.background_image, QRectF(self.background_image.rect()))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that creates a status widgets which can be used the state of an app
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Property, QSize, QTimer
from Qt.QtWidgets import QSizePolicy, QHBoxLayout, QFrame
from Qt.QtGui import QPixmap

from tpDcc.managers import resources
from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.widgets import label


@theme.mixin
class StatusWidget(QFrame, object):

    DEFAULT_DISPLAY_TIME = 10000  # milliseconds -> 10 seconds

    def __init__(self, *args):
        super(StatusWidget, self).__init__(*args)

        self._status = None
        self._blocking = False
        self._timer = QTimer(self)

        self.setObjectName('StatusWidget')
        self.setFrameShape(QFrame.NoFrame)
        self.setFixedHeight(19)
        self.setMinimumWidth(5)

        self._label = label.BaseLabel('', parent=self)
        self._label.setStyleSheet('background-color: transparent;')
        self._label.setCursor(Qt.IBeamCursor)
        self._label.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self._label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        self.label_image = label.BaseLabel(parent=self)
        self.label_image.setMaximumSize(QSize(17, 17))
        self.label_image.hide()

        self.main_layout = QHBoxLayout(self)
        self.main_layout.setContentsMargins(1, 0, 0, 0)

        self.main_layout.addWidget(self.label_image)
        self.main_layout.addWidget(self._label)

        self.setLayout(self.main_layout)

        self._timer.timeout.connect(self._reset)

        # Force set to initialize default status Qt property
        self.status = ''

    def _get_status(self):
        return self._status

    def _set_status(self, value):
        self._status = str(value)
        self.polish()

    status = Property(str, _get_status, _set_status)

    def is_blocking(self):
        """
        Returns True if the status widget is blocking, otherwise return False
        :return: bool
        """

        return self._blocking

    def show_ok_message(self, message, msecs=None):
        """
        Set an ok message to be displayed in the status widget
        :param message: str
        :param msecs: int
        """

        if self.is_blocking():
            return

        self.status = 'ok'
        icon = resources.icon('ok')
        self._show_message(message, icon, msecs)

    def show_info_message(self, message, msecs=None):
        """
        Set an info message to be displayed in the status widget
        :param message: str
        :param msecs: int
        """

        if self.is_blocking():
            return

        self.status = 'info'
        icon = resources.icon('info')
        self._show_message(message, icon, msecs)

    def show_warning_message(self, message, msecs=None):
        """
       Set a warning message to be displayed in the status widget
       :param message: str
       :param msecs: int
       """

        if self.is_blocking():
            return

        self.status = 'warning'
        icon = resources.icon('warning')
        self._show_message(message, icon, msecs)

    def show_error_message(self, message, msecs=None):
        """
       Set an error message to be displayed in the status widget
       :param message: str
       :param msecs: int
       """

        self.status = 'error'
        icon = resources.icon('error', extension='png')
        self._show_message(message, icon, msecs, blocking=True)

    def _reset(self):
        """
        Called when the current animation has finished
        """

        self._timer.stop()
        self.label_image.setVisible(False)
        self._label.setText('')
        icon = resources.pixmap('blank')
        self.label_image.setPixmap(icon) if icon else self.label_image.setPixmap(QPixmap())
        self.setStyleSheet('')
        self._blocking = False
        self.status = ''

    def _show_message(self, message, icon, msecs=None, blocking=False):
        """
        Set the given text to be displayed in the status widget
        :param message: str
        :param icon: QIcon
        :param msecs: int
        :param blocking: bool
        """

        msecs = msecs or self.DEFAULT_DISPLAY_TIME
        self._blocking = blocking

        self.label_image.setStyleSheet('border: 0px;')

        if icon:
            self.label_image.setPixmap(icon.pixmap(QSize(17, 17)))
            self.label_image.show()
        else:
            self.label_image.hide()

        if message:
            self._label.setText(str(message))
            self._timer.stop()
            self._timer.start(msecs)
        else:
            self._reset()

        self.update()



import tpDcc
from tpDcc.libs.qt.core import base, mixin
from tpDcc.libs.qt.widgets import layouts, dividers, spinbox, datetime


class SpinBoxExample(tpDcc.Window, mixin.FieldMixin):
    def __init__(self):
        super(SpinBoxExample, self).__init__(title='Example for SpinBoxes')

    def ui(self):
        super(SpinBoxExample, self).ui()

        class_list = [
            spinbox.BaseSpinBox, spinbox.BaseDoubleSpinBox,
            datetime.BaseDateTimeEdit, datetime.BaseDateEdit, datetime.BaseTimeEdit
        ]

        for cls in class_list:
            self.main_layout.addWidget(dividers.Divider(cls.__name__))
            lyt = layouts.HorizontalLayout()
            lyt.addWidget(cls().small())
            lyt.addWidget(cls().medium())
            lyt.addWidget(cls().large())
            self.main_layout.addLayout(lyt)

        extra_spinners_lyt1 = layouts.HorizontalLayout()
        spn1 = spinbox.BaseSpinBoxNumber(parent=self)
        spn2 = spinbox.BaseDoubleNumberSpinBox(parent=self)

        extra_spinners_lyt2 = layouts.HorizontalLayout()
        spn3 = spinbox.DragDoubleSpinBox(parent=self)
        spn4 = spinbox.DragDoubleSpinBoxLine(parent=self)

        extra_spinners_lyt1.addWidget(spn1)
        extra_spinners_lyt1.addWidget(spn2)
        extra_spinners_lyt2.addWidget(spn3)
        extra_spinners_lyt2.addWidget(spn4)

        self.main_layout.addWidget(dividers.Divider('Extra Base'))
        self.main_layout.addLayout(extra_spinners_lyt1)

        self.main_layout.addWidget(dividers.Divider('Draggable'))
        self.main_layout.addLayout(extra_spinners_lyt2)

        self.main_layout.addWidget(dividers.Divider('Pop calendar widget'))
        date_time_edit = datetime.BaseDateTimeEdit()
        date_time_edit.setCalendarPopup(True)
        date_edit = datetime.BaseDateEdit()
        date_edit.setCalendarPopup(True)
        time_edit = datetime.BaseTimeEdit()
        time_edit.setCalendarPopup(True)
        date_lyt = layouts.HorizontalLayout()
        date_lyt.addWidget(date_time_edit)
        date_lyt.addWidget(time_edit)
        date_lyt.addWidget(date_edit)
        self.main_layout.addLayout(date_lyt)
        self.main_layout.addStretch()


def run():
    return SpinBoxExample()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains code/script related widgets
"""

from __future__ import print_function, division, absolute_import

import re
import sys
import string
import logging

from Qt.QtCore import Qt, Signal, QRect, QSize, QRegExp, QStringListModel, QFile
from Qt.QtWidgets import QWidget, QCompleter, QTextEdit, QPlainTextEdit, QShortcut
from Qt.QtGui import QFont, QColor, QPainter, QTextCursor, QTextCharFormat, QTextOption, QTextFormat, QSyntaxHighlighter
from Qt.QtGui import QKeySequence

from tpDcc import dcc
from tpDcc.libs.python import python, fileio, code, folder as folder_utils, path as path_utils
from tpDcc.libs.qt.core import qtutils

LOGGER = logging.getLogger('tpDcc-libs-qt')


class PythonCompleter(QCompleter, object):
    def __init__(self):
        super(PythonCompleter, self).__init__()

        self._info = None
        self._file_path = None
        self._model_strings = list()
        self._reset_list = True
        self._string_model = QStringListModel(self._model_strings, self)

        self._refresh_completer = True
        self._sub_activated = False
        self._last_imports = None
        self._last_lines = None
        self._last_path = None
        self._current_defined_imports = None
        self._last_path_and_part = None
        self._current_sub_functions = None
        self._last_column = 0

        self.setCompletionMode(self.PopupCompletion)
        self.setCaseSensitivity(Qt.CaseInsensitive)
        self.setModel(self._string_model)
        self.setWrapAround(False)

        self.activated.connect(self._on_insert_completion)

    def setWidget(self, widget):
        super(PythonCompleter, self).setWidget(widget)
        self.setParent(widget)

    def keyPressEvent(self):
        return

    def show_info_popup(self, info=None):
        self._info = QTextEdit()
        self._info.setEnabled(False)
        self._info.setWindowFlags(Qt.Popup)
        self._info.show()

    def get_imports(self, paths=None):
        imports = self._get_available_modules(paths=paths)
        imports.sort()

        return imports

    def get_sub_imports(self, path):
        """
        Returns namespace in a module
        :param path: str
        :return: str
        """

        defined = code.get_defined(path)
        defined.sort()

        return defined

    def clear_completer_list(self):
        self._string_model.setStringList([])

    def text_under_cursor(self):
        cursor = self.widget().textCursor()
        cursor.select(cursor.LineUnderCursor)

        return cursor.selectedText()

    def set_filepath(self, file_path):
        if not file_path:
            return

        self._file_path = file_path

    def handle_text(self, text):
        """
        Parses a single line of text
        :param text: str
        :return: bool
        """

        if not text:
            return False

        cursor = self.widget().textCursor()
        column = cursor.columnNumber() - 1
        if column < self._last_column:
            self._last_column = column
            return False
        self._last_column = column

        if column == 1:
            return False
        text = str(text)
        passed = self.handle_from_import(text, column)
        if passed:
            return True
        passed = self.handle_sub_import(text, column)
        if passed:
            return True
        passed = self.handle_import_load(text, cursor)
        if passed:
            return True

        return False

    def handle_import(self, text):
        m = re.search(r'(from|import)(?:\s+?)(\w*)', text)
        if m:
            # TODO: Find available modules in Python path
            pass

    def handle_sub_import(self, text, column):
        m = re.search(r'(from|import)(?:\s+?)(\w*.?\w*)\.(\w*)$', text)
        if m:
            if column < m.end(2):
                return False
            from_module = m.group(2)
            module_path = code.get_package_path_from_name(from_module)
            last_part = m.group(3)
            if module_path:
                defined = self.get_imports(module_path)
                self._string_model.setStringList(defined)
                self.setCompletionPrefix(last_part)
                self.popup().setCurrentIndex(self.completionModel().index(0, 0))
                return True

        return False

    def handle_import_load(self, text, cursor):
        column = cursor.columnNumber() - 1
        text = text[:cursor.columnNumber()]
        m = re.search(r'\s*([a-zA-Z0-9._]+)\.([a-zA-Z0-9_]*)$', text)
        block_number = cursor.blockNumber()
        line_number = block_number + 1
        all_text = self.widget().toPlainText()
        scope_text = all_text[:(cursor.position() - 1)]

        if m and m.group(2):
            scope_text = all_text[:(cursor.position() - len(m.group(2)) + 1)]
        if not m:
            return False

        assignment = m.group(1)
        if column < m.end(1):
            return False

        sub_m = re.search(r'(from|import)\s+(%s)' % assignment, text)
        if sub_m:
            return False

        path = None
        sub_part = None
        target = None

        text = self.widget().toPlainText()
        lines = fileio.get_text_lines(text)

        # Search for assignments
        assign_map = code.get_ast_assignment(scope_text, line_number - 1, assignment)
        if assign_map:
            if assignment in assign_map:
                target = assign_map[assignment]
            else:
                split_assignment = assignment.split('.')
                inc = 1

                while assignment not in assign_map:
                    sub_assignment = string.join(split_assignment[:(inc * -1)], '.')
                    if sub_assignment in assign_map:
                        target = assign_map[sub_assignment]
                        break
                    inc += 1
                    if inc > (len(split_assignment) - 1):
                        break
                sub_part = string.join(split_assignment[inc:], '.')

        module_name = m.group(1)
        if target and len(target) == 2:
            if target[0] == 'import':
                module_name = target[1]
            if not target[0] == 'import':
                module_name = target[0]
                sub_part = target[1]

        # import from module
        if module_name:
            imports = None
            if lines == self.last_lines:
                imports = self.last_imports
            if not imports:
                imports = code.get_line_imports(lines)

            self._last_imports = imports
            self._last_lines = lines

            if module_name in imports:
                path = imports[module_name]
            if module_name not in imports:
                split_assignment = module_name.split('.')
                last_part = split_assignment[-1]
                if last_part in imports:
                    path = imports[last_part]

            if path and not sub_part:
                test_text = ''
                defined = None
                if path == self.last_path:
                    defined = self.current_defined_imports
                if len(m.groups()) > 0:
                    test_text = m.group(2)
                if not defined:
                    defined = self.get_imports(path)
                    if defined:
                        self._current_defined_imports = defined
                    else:
                        defined = self.get_sub_imports(path)

                custom_defined = self.custom_import_load(assign_map, module_name)
                if custom_defined:
                    defined = custom_defined
                if defined:
                    if test_text and test_text[0].islower():
                        defined.sort(key=str.swapcase)
                    self._string_model.setStringList(defined)
                    self.setCompletionPrefix(test_text)
                    self.setCaseSensitivity(Qt.CaseInsensitive)
                    self.popup().setCurrentIndex(self.completionModel().index(0, 0))
                    return True

            # import from a class of a module
            if path and sub_part:

                sub_functions = None
                if self.last_path_and_part:
                    if path == self.last_path_and_part[0] and sub_part == self.last_path_and_part[1]:
                        sub_functions = self.current_sub_functions

                if not sub_functions:
                    sub_functions = code.get_ast_class_sub_functions(path, sub_part)
                    if sub_functions:
                        self._current_sub_functions = sub_functions

                self._last_path_and_part = [path, sub_part]
                if not sub_functions:
                    return False

                test_text = ''
                if len(m.groups()) > 0:
                    test_text = m.group(2)
                if test_text and test_text[0].islower():
                    sub_functions.sort(key=str.swapcase)
                self._string_model.setStringList(sub_functions)
                self.setCompletionPrefix(test_text)
                self.setCaseSensitivity(Qt.CaseInsensitive)
                self.popup().setCurrentIndex(self.completionModel().index(0, 0))
                return True

        module_name = m.group(1)
        if module_name:
            custom_defined = self.custom_import_load(assign_map, module_name)

            test_text = ''
            if len(m.groups()) > 0:
                test_text = m.group(2)

            if test_text and test_text[0].islower():
                custom_defined.sort(key=str.swapcase)
            self._string_model.setStringList(custom_defined)
            self.setCompletionPrefix(test_text)
            self.setCaseSensitivity(Qt.CaseInsensitive)
            self.popup().setCurrentIndex(self.completionModel().index(0, 0))
            return True

        return False

    def handle_from_import(self, text, column):
        m = re.search(r'(from)(?:\s+?)(\w*.?\w*)(?:\s+?)(import)(?:\s+?)(\w+)?$', text)
        if m:
            if column < m.end(3):
                return False
            from_module = m.group(2)
            module_path = code.get_package_path_from_name(from_module)

            last_part = m.group(4)
            if not last_part:
                last_part = ''
            if module_path:
                defined = self.get_imports(module_path)
                self._string_model.setStringList(defined)
                self.setCompletionPrefix(last_part)
                self.popup().setCurrentIndex(self.completionModel().index(0, 0))

                return True

        return False

    def custom_import_load(self, assign_map, moduel_name):
        return

    def _get_available_modules(self, paths=None):
        imports = list()
        if not paths:
            paths = sys.path
        if paths:
            paths = python.force_list(paths)

        for path in paths:
            fix_path = path_utils.normalize_path(path)
            stuff_in_folder = folder_utils.get_files_and_folders(fix_path)
            for file_or_folder in stuff_in_folder:
                folder_path = path_utils.join_path(fix_path, file_or_folder)
                files = folder_utils.get_files_with_extension('py', folder_path, full_path=False)
                if '__init__.py' in files:
                    imports.append(str(file_or_folder))

            python_files = folder_utils.get_files_with_extension('py', fix_path, full_path=False)
            for python_file in python_files:
                if python_file.startswith('__'):
                    continue
                python_file_name = python_file.split('.')[0]
                imports.append(str(python_file_name))

        if imports:
            imports = list(set(imports))

        return imports

    def _on_insert_completion(self, completion_string):
        widget = self.widget()
        cursor = widget.textCursor()
        if completion_string == self.completionPrefix():
            return
        extra = len(self.completionPrefix())
        cursor.movePosition(QTextCursor.Left, cursor.KeepAnchor, extra)
        cursor.removeSelectedText()
        cursor.insertText(completion_string)
        widget.setTextCursor(cursor)


def get_syntax_format(color=None, style=''):
    """
    Returns a QTextCharFormat with the given attributes.
    """

    _color = None

    if type(color) == str:
        _color = QColor()
        _color.setNamedColor(color)
    if type(color) == list:
        _color = QColor(*color)

    if color == 'green':
        _color = Qt.green

    _format = QTextCharFormat()

    if _color:
        _format.setForeground(_color)
    if 'bold' in style:
        _format.setFontWeight(QFont.Bold)
    if 'italic' in style:
        _format.setFontItalic(True)

    return _format


def syntax_styles(name):
    if name == 'keyword':
        if dcc.is_maya():
            return get_syntax_format('green', 'bold')
        if not dcc.is_maya():
            return get_syntax_format([0, 150, 150], 'bold')
    if name == 'operator':
        if dcc.is_maya():
            return get_syntax_format('gray')
        if not dcc.is_maya():
            return get_syntax_format('darkGray')
    if name == 'brace':
        if dcc.is_maya():
            return get_syntax_format('lightGray')
        if not dcc.is_maya():
            return get_syntax_format('darkGray')
    if name == 'defclass':
        if dcc.is_maya():
            return get_syntax_format(None, 'bold')
        if not dcc.is_maya():
            return get_syntax_format(None, 'bold')
    if name == 'string':
        if dcc.is_maya():
            return get_syntax_format([230, 230, 0])
        if not dcc.is_maya():
            return get_syntax_format('blue')
    if name == 'string2':
        if dcc.is_maya():
            return get_syntax_format([230, 230, 0])
        if not dcc.is_maya():
            return get_syntax_format('lightGreen')
    if name == 'comment':
        if dcc.is_maya():
            return get_syntax_format('red')
        if not dcc.is_maya():
            return get_syntax_format('red')
    if name == 'self':
        if dcc.is_maya():
            return get_syntax_format(None, 'italic')
        if not dcc.is_maya():
            return get_syntax_format('black', 'italic')
    if name == 'bold':
        return get_syntax_format(None, 'bold')
    if name == 'numbers':
        if dcc.is_maya():
            return get_syntax_format('cyan')
        if not dcc.is_maya():
            return get_syntax_format('brown')


class PythonHighlighter(QSyntaxHighlighter):
    """
    Syntax highlighter for the Python language.
    """

    # Python keywords
    keywords = [
        'and', 'assert', 'break', 'class', 'continue', 'def',
        'del', 'elif', 'else', 'except', 'exec', 'finally',
        'for', 'from', 'global', 'if', 'import', 'in',
        'is', 'lambda', 'not', 'or', 'pass', 'print',
        'raise', 'return', 'try', 'while', 'yield',
        'None', 'True', 'False', 'process', 'show'
    ]

    if dcc.is_maya():
        keywords += ['cmds', 'pm', 'mc', 'pymel']

    # Python operators
    operators = [
        '=',
        # Comparison
        '==', '!=', '<', '<=', '>', '>=',
        # Arithmetic
        r'\+', '-', r'\*', '/', '//', r'\%', r'\*\*',
        # In-place
        r'\+=', '-=', r'\*=', '/=', r'\%=',
        # Bitwise
        r'\^', r'\|', r'\&', r'\~', '>>', '<<',
    ]

    # Python braces
    braces = [
        r'\{', r'\}', r'\(', r'\)', r'\[', r'\]',
    ]

    def __init__(self, document):
        super(PythonHighlighter, self).__init__(document)

        # Multi-line strings (expression, flag, style)
        # FIXME: The triple-quotes in these two lines will mess up the
        # syntax highlighting from this point onward
        self.tri_single = (QRegExp("'''"), 1, syntax_styles('string2'))
        self.tri_double = (QRegExp('"""'), 2, syntax_styles('string2'))

        rules = []

        # Keyword, operator, and brace rules
        rules += [(r'\b%s\b' % w, 0, syntax_styles('keyword'))
                  for w in PythonHighlighter.keywords]
        rules += [(r'%s' % o, 0, syntax_styles('operator'))
                  for o in PythonHighlighter.operators]
        rules += [(r'%s' % b, 0, syntax_styles('brace'))
                  for b in PythonHighlighter.braces]

        # All other rules
        rules += [
            # 'self'
            (r'\bself\b', 0, syntax_styles('self')),

            # Double-quoted string, possibly containing escape sequences
            (r'"[^"\\]*(\\.[^"\\]*)*"', 0, syntax_styles('string')),
            # Single-quoted string, possibly containing escape sequences
            (r"'[^'\\]*(\\.[^'\\]*)*'", 0, syntax_styles('string')),

            # 'def' followed by an identifier
            # (r'\bdef\b\s*(\w+)', 0, syntax_styles('defclass')),
            # 'class' followed by an identifier
            # (r'\bclass\b\s*(\w+)', 0, syntax_styles('defclass')),

            # From '#' until a newline
            (r'#[^\n]*', 0, syntax_styles('comment')),
            # ('\\b\.[a-zA-Z_]+\\b(?=\()', 0, syntax_styles('bold')),
            # Numeric literals
            (r'\b[+-]?[0-9]+[lL]?\b', 0, syntax_styles('numbers')),
            (r'\b[+-]?0[xX][0-9A-Fa-f]+[lL]?\b', 0, syntax_styles('numbers')),
            (r'\b[+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?\b', 0, syntax_styles('numbers')),
        ]

        # Build a QRegExp for each pattern
        self.rules = [(QRegExp(pat), index, fmt) for (pat, index, fmt) in rules]

    def highlightBlock(self, text):
        """
        Apply syntax highlighting to the given block of text.
        """

        # Do other syntax formatting
        for expression, nth, format_value in self.rules:
            index = expression.indexIn(text, 0)

            while index >= 0:
                # We actually want the index of the nth match
                index = expression.pos(nth)
                length = len(expression.cap(nth))
                self.setFormat(index, length, format_value)
                index = expression.indexIn(text, index + length)

        self.setCurrentBlockState(0)

        # Do multi-line strings
        in_multiline = self.match_multiline(text, *self.tri_single)
        if not in_multiline:
            in_multiline = self.match_multiline(text, *self.tri_double)

    def match_multiline(self, text, delimiter, in_state, style):
        """
        Do highlighting of multi-line strings. ``delimiter`` should be a
        ``QRegExp`` for triple-single-quotes or triple-double-quotes, and
        ``in_state`` should be a unique integer to represent the corresponding
        state changes when inside those strings. Returns True if we're still
        inside a multi-line string when this function is finished.
        """

        # If inside triple-single quotes, start at 0
        if self.previousBlockState() == in_state:
            start = 0
            add = 0
        # Otherwise, look for the delimiter on this line
        else:
            start = delimiter.indexIn(text)
            # Move past this match
            add = delimiter.matchedLength()

        # As long as there's a delimiter match on this line...
        while start >= 0:
            # Look for the ending delimiter
            end = delimiter.indexIn(text, start + add)
            # Ending delimiter on this line?
            if end >= add:
                length = end - start + add + delimiter.matchedLength()
                self.setCurrentBlockState(0)
            # No; multi-line string
            else:
                self.setCurrentBlockState(in_state)
                length = len(text) - start + add
            # Apply formatting
            self.setFormat(start, length, style)
            # Look for the next match
            start = delimiter.indexIn(text, start + length)

        # Return True if still inside a multi-line string, False otherwise
        if self.currentBlockState() == in_state:
            return True
        else:
            return False


class CodeLineNumber(QWidget, object):
    def __init__(self, code_editor):
        super(CodeLineNumber, self).__init__()

        self.setParent(code_editor)
        self._code_editor = code_editor

    def sizeHint(self):
        return QSize(self._code_editor._line_number_width(), 0)

    def paintEvent(self, event):
        self._code_editor._line_number_paint(event)


class CodeCompleter(PythonCompleter, object):
    def __init__(self):
        super(CodeCompleter, self).__init__()

    def keyPressEvent(self):
        return

    def _on_insert_completion(self, completion_string):
        super(CodeCompleter, self)._on_insert_completion(completion_string)

        # This stops Maya from entering edit mode in the outliner, if something is selected
        if dcc.is_maya():
            dcc.focus('modelPanel1')

    def _format_live_function(self, function_instance):
        function_name = None
        if hasattr(function_instance, 'im_func'):
            args = function_instance.im_func.func_code.co_varnames
            count = function_instance.im_func.func_code.co_argcount
            args_name = ''
            if args:
                args = args[:count]
                if args[0] == 'self':
                    args = args[1:]
                args_name = string.join(args, ',')
            function_name = '{}({})'.format(function_instance.im_func.func.name, args_name)

        return function_name

    def custom_import_load(self, assign_map, module_name):

        found = list()

        if module_name == 'cmds':
            if assign_map:
                if module_name in assign_map:
                    return []
            if dcc.is_maya():
                import maya.cmds as cmds
                functions = dir(cmds)
                return functions

        return found


class CodeTextEdit(QPlainTextEdit, object):

    save = Signal(object)
    saveDone = Signal(object)
    fileSet = Signal()
    findOpened = Signal(object)
    codeTextSizeChanged = Signal(object)
    mousePressed = Signal(object)

    def __init__(self, settings=None, parent=None):

        self._settings = settings
        self._text_edit_globals = dict()
        self._file_path = None
        self._last_modified = None
        self._find_widget = None
        self._completer = None
        self._skip_focus = False

        self._option_object = None

        super(CodeTextEdit, self).__init__(parent=parent)

        self.setFont(QFont('Courier', 9))
        self.setWordWrapMode(QTextOption.NoWrap)

        save_shortcut = QShortcut(QKeySequence(self.tr('Ctlr+s')), self)
        find_shortcut = QShortcut(QKeySequence(self.tr('Ctrl+f')), self)
        goto_line_shortcut = QShortcut(QKeySequence(self.tr('Ctrl+g')), self)
        duplicate_line_shortcut = QShortcut(QKeySequence(self.tr('Ctrl+d')), self)
        zoom_in_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Plus), self)
        zoom_in_other_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Equal), self)
        zoom_out_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Minus), self)

        save_shortcut.activated.connect(self._on_save)
        find_shortcut.activated.connect(self._on_find)
        goto_line_shortcut.activated.connect(self._on_goto_line)
        duplicate_line_shortcut.activated.connect(self._on_duplicate_line)
        zoom_in_shortcut.activated.connect(self._zoom_in_text)
        zoom_in_other_shortcut.activated.connect(self._zoom_in_text)
        zoom_out_shortcut.activated.connect(self._zoom_out_text)

        self._line_numbers = CodeLineNumber(self)

        self._setup_highlighter()
        self._update_number_width(0)
        self._line_number_highlight()

        self.blockCountChanged.connect(self._on_update_number_width)
        self.updateRequest.connect(self._on_update_number_area)
        self.cursorPositionChanged.connect(self._on_cursor_position_changed)
        self.codeTextSizeChanged.connect(self._on_code_text_size_changed)

    def resizeEvent(self, event):
        super(CodeTextEdit, self).resizeEvent(event)
        rect = self.contentsRect()
        new_rect = QRect(rect.left(), rect.top(), self._line_number_width(), rect.height())
        self._line_numbers.setGeometry(new_rect)

    def mousePressEvent(self, event):
        self.mousePressed.emit(event)
        return super(CodeTextEdit, self).mousePressEvent(event)

    def keyPressEvent(self, event):

        pass_on = True
        quit_right_away = False

        if event.key() in (Qt.Key_Right, Qt.Key_Left, Qt.Key_Up, Qt.Key_Down):
            quit_right_away = True
        if quit_right_away:
            super(CodeTextEdit, self).keyPressEvent(event)
            return

        if self._completer:
            self._completer.activated.connect(self._on_activate)

            if self._completer.popup().isVisible():
                if event.key() == Qt.Key_Enter:
                    event.ignore()
                    return
                elif event.key() == Qt.Key_Return:
                    event.ignore()
                    return
                elif event.key() == Qt.Key_Escape:
                    event.ignore()
                    return
                elif event.key() == Qt.Key_Tab:
                    event.ignore()
                    return
                elif event.key() == Qt.Key_Backtab:
                    event.ignore()
                    return
            else:
                if event.key() == Qt.Key_Control or event.key() == Qt.Key_Shift:
                    event.ignore()
                    self._completer.popup().hide()
                    return
            if event.key() == Qt.Key_Control:
                event.ignore()
                self._completer.popup().hide()
                return

        if event.modifiers() and Qt.ControlModifier:
            if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
                self._run()
                return
        if event.key() == Qt.Key_Backtab or event.key() == Qt.Key_Tab:
            self._handle_tab(event)
            pass_on = False
        if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
            self._handle_enter(event)
            pass_on = False

        if pass_on:
            super(CodeTextEdit, self).keyPressEvent(event)

        if self._completer:
            text = self._completer.text_under_cursor()
            if text:
                result = self._completer.handle_text(text)
                if result:
                    rect = self.cursorRect()
                    scroll_width = self._completer.popup().verticalScrollBar().sizeHint().width()
                    width = self._completer.popup().sizeHintForColumn(0) + scroll_width
                    if width > 350:
                        width = 350
                    rect.setWidth(width)
                    self._completer.complete(rect)
                if not result:
                    self._completer.popup().hide()
                    self._completer.clear_completer_list()
                    self._completer.refresh_completer = True

    def wheelEvent(self, event):
        delta = event.delta()
        keys = event.modifiers()
        if keys == Qt.CTRL:
            if delta > 0:
                self._zoom_in_text()
            elif delta < 0:
                self._zoom_out_text()

        return super(CodeTextEdit, self).wheelEvent(event)

    def focusInEvent(self, event):
        if self._completer:
            self._completer.setWidget(self)
        super(CodeTextEdit, self).focusInEvent(event)
        if not self._skip_focus:
            self._update_request()

    def get_settings(self):
        return self._settings

    def set_settigns(self, settings):
        self._settings = settings

    def set_option_object(self, option_object):
        self._option_object = option_object

    def set_completer(self, completer):
        self._completer = completer()
        self._completer.setWidget(self)
        self._completer.set_filepath(self._file_path)

    def set_file(self, file_path):
        in_file = QFile(file_path)
        if in_file.open(QFile.ReadOnly | QFile.Text):
            text = in_file.readAll()
            self.setPlainText(text)
        self._file_path = file_path
        self._last_modified = fileio.get_last_modified_date(self._file_path)
        if self._completer:
            self._completer.set_filepath(file_path)
        self.fileSet.emit()

    def set_find_widget(self, widget):
        self._find_widget.set_widget(widget)

    def load_modification_date(self):
        self._last_modified = fileio.get_last_modified_date(self._file_path)

    def is_modified(self):
        return self.document().isModified()

    def _handle_enter(self, event):
        cursor = self.textCursor()
        current_block = cursor.block()
        cursor_position = cursor.positionInBlock()
        current_block_text = str(current_block.text())

        current_found = ''
        if not current_found:
            current_found = re.search('^ +', current_block_text)
            if current_found:
                current_found = current_found.group(0)

        indent = 0
        if current_found:
            indent = len(current_found)

        colon_position = current_block_text.find(':')
        comment_position = current_block_text.find('#')
        if colon_position > -1:
            sub_indent = 4
            if -1 < comment_position < colon_position:
                sub_indent = 0
            indent += sub_indent
        if cursor_position < indent:
            indent = (cursor_position - indent) + indent

        cursor.insertText(('\n' + ' ' * indent))

    def _handle_tab(self, event):
        cursor = self.textCursor()
        document = self.document()

        start_position = cursor.anchor()
        select_position = cursor.selectionStart()
        select_start_block = document.findBlock(select_position)
        start = select_position - select_start_block.position()
        end_position = cursor.position()
        if start_position > end_position:
            temp_position = end_position
            end_position = start_position
            start_position = temp_position

        if event.key() == Qt.Key_Tab:
            if not cursor.hasSelection():
                self.insertPlainText('    ')
                start_position += 4
                end_position = start_position
            else:
                cursor.setPosition(start_position)
                cursor.movePosition(QTextCursor.StartOfLine)
                cursor.setPosition(end_position, QTextCursor.KeepAnchor)
                text = cursor.selection().toPlainText()
                split_text = text.split('\n')
                edited = list()
                index = 0
                for text_split in split_text:
                    edited.append(self._add_tab(text_split))
                    if index == 0:
                        start_position += 4
                    end_position += 4
                    index += 1

                edited_text = string.join(edited, '\n')
                cursor.insertText(edited_text)
                self.setTextCursor(cursor)

        elif event.key() == Qt.Key_Backtab:
            if not cursor.hasSelection():
                cursor = self.textCursor()
                cursor.movePosition(QTextCursor.StartOfLine)
                cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, 4)
                text = str(cursor.selection().toPlainText())
                if text and text == '    ':
                    cursor.insertText('')
                    self.setTextCursor(cursor)
                    start_position -= 4
                    end_position = start_position
            else:
                cursor.setPosition(start_position)
                cursor.movePosition(QTextCursor.StartOfLine)
                cursor.setPosition(end_position, QTextCursor.KeepAnchor)
                cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
                self.setTextCursor(cursor)
                text = str(cursor.selection().toPlainText())
                split_text = text.split('\n')
                edited = list()
                index = 0
                skip_indent = False
                for text_split in split_text:
                    new_string_value = text_split
                    if not skip_indent:
                        new_string_value = self._remove_tab(text_split)
                    if index == 0 and new_string_value == text_split:
                        skip_indent = True
                    if not skip_indent:
                        if new_string_value != text_split:
                            if index == 0:
                                offset = (start - 4) + 4
                                if offset > 4:
                                    offset = 4
                                start_position -= offset
                            end_position -= 4
                    edited.append(new_string_value)
                    index += 1

                edited_text = string.join(edited, '\n')
                cursor.insertText(edited_text)
                self.setTextCursor(cursor)

        cursor = self.textCursor()
        cursor.setPosition(start_position)
        cursor.setPosition(end_position, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def _add_tab(self, string_value):
        return '    {}'.format(string_value)

    def _remove_tab(self, string_value):
        string_section = string_value[0:4]
        if string_section == '    ':
            return string_value[4:]

        return string_value

    def _set_text_size(self, value):
        font = self.font()
        font.setPixelSize(value)
        self.setFont(font)

    def _code_text_size_change(self, value):
        self._set_text_size(value)

    def _setup_highlighter(self):
        self._higlighter = PythonHighlighter(document=self.document())

    def _line_number_paint(self, event):
        paint = QPainter(self._line_numbers)
        if not dcc.is_maya():
            paint.fillRect(event.rect(), Qt.lightGray)
        else:
            paint.fillRect(event.rect(), Qt.black)

        block = self.firstVisibleBlock()
        block_number = block.blockNumber()
        top = int(self.blockBoundingGeometry(block).translated(self.contentOffset()).top())
        bottom = int(top + self.blockBoundingGeometry(block).height())
        while block.isValid() and top <= event.rect().bottom():
            if block.isVisible() and bottom >= event.rect().top():
                number = block_number + 1
                if dcc.is_maya():
                    paint.setPen(Qt.lightGray)
                else:
                    paint.setPen(Qt.black)
                paint.drawText(
                    0, top, self._line_numbers.width(), self.fontMetrics().height(), Qt.AlignRight, str(number))
            block = block.next()
            top = bottom
            bottom = top + self.blockBoundingRect(block).height()
            block_number += 1

    def _line_number_width(self):
        digits = 1
        max_value = max(1, self.blockCount())
        while max_value >= 10:
            max_value /= 10
            digits += 1
        space = 1 + self.fontMetrics().width('1') * digits

        return space

    def _line_number_highlight(self):
        extra_selection = QTextEdit.ExtraSelection()
        selections = [extra_selection]
        if not self.isReadOnly():
            selection = QTextEdit.ExtraSelection()
            if dcc.is_maya():
                line_color = QColor(Qt.black)
            else:
                line_color = QColor(Qt.lightGray)
            selection.format.setBackground(line_color)
            selection.format.setProperty(QTextFormat.FullWidthSelection, True)
            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            selections.append(selection)

        self.setExtraSelections(selections)

    def _update_number_width(self, value=0):
        self.setViewportMargins(self._line_number_width(), 0, 0, 0)

    def _update_number_area(self, rect, y_value):
        if y_value:
            self._line_numbers.scroll(0, y_value)
        else:
            self._line_numbers.update(0, rect.y(), self._line_number_width(), rect.height())
        if rect.contains(self.viewport().rect()):
            self._update_number_width()

    def _has_changed(self):
        if self._file_path:
            if path_utils.is_file(self._file_path):
                last_modified = fileio.get_last_modified_date(self._file_path)
                if last_modified != self._last_modified:
                    return True

        return False

    def _update_request(self):
        if not self._has_changed():
            return

        last_modified = fileio.get_last_modified_date(self._file_path)
        self._skip_focus = True
        permission = qtutils.get_permission(
            'File:\n{}\nhas changed, do you want to relaod it?'.format(path_utils.get_basename(self._file_path)), self)
        if permission:
            self.set_file(self._file_path)
        else:
            self._last_modified = last_modified

        self._skip_focus = False

    def _zoom_in_text(self):
        font = self.font()
        size = font.pointSize()
        size += 1
        font.setPointSize(size)
        self.setFont(QFont('Courier', size))

    def _zoom_out_text(self):
        font = self.font()
        size = font.pointSize()
        size -= 1
        font.setPointSize(size)
        self.setFont(QFont('Courier', size))

    def _run(self):
        raise NotImplementedError('run functionality is not implemented!')

    def _on_save(self):
        if not self.document().isModified():
            LOGGER.warning('No changes to save in {}'.format(self._file_path))
            return

        old_last_modified = self._last_modified
        try:
            self.save.emit(self)
        except Exception:
            pass

        new_last_modified = fileio.get_last_modified_date(self._file_path)
        if old_last_modified == new_last_modified:
            self.saveDone.emit(False)

        if old_last_modified != new_last_modified:
            self.saveDone.emit(True)
            self.document().setModified(False)
            self._last_modified = new_last_modified

    def _on_find(self):
        self.findOpened.emit(self)

    def _on_goto_line(self):
        line = qtutils.get_comment(parent=self, text_message='', title='Goto Line')
        if not line:
            return

        line_number = int(line)
        text_cursor = self.textCursor()
        block_number = text_cursor.blockNumber()

        number = line_number - block_number
        if number > 0:
            move_type = text_cursor.NextBlock
            number -= 2
        if number < 0:
            move_type = text_cursor.PreviousBlock
            number = abs(number)

        text_cursor.movePosition(move_type, text_cursor.MoveAnchor, (number + 1))
        self.setTextCursor(text_cursor)

    def _on_duplicate_line(self):
        text_cursor = self.textCursor()

        selected_text = text_cursor.selectedText()
        if not selected_text:
            text_cursor.beginEditBlock()
            text_cursor.movePosition(QTextCursor.StartOfLine, QTextCursor.MoveAnchor)
            text_cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
            selected_text = text_cursor.selectedText()
            text_cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.MoveAnchor)
            text_cursor.insertBlock()
            text_cursor.insertText(selected_text)
            text_cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.MoveAnchor)
            self.setTextCursor(text_cursor)
            text_cursor.endEditBlock()
        else:
            text_cursor.beginEditBlock()
            text_cursor.setPosition(text_cursor.selectionEnd(), QTextCursor.MoveAnchor)
            self.setTextCursor(text_cursor)
            text_cursor.insertBlock()
            position = text_cursor.position()
            text_cursor.insertText(selected_text)
            end_position = text_cursor.position()
            text_cursor.setPosition(position, QTextCursor.MoveAnchor)
            text_cursor.setPosition(end_position, QTextCursor.KeepAnchor)
            self.setTextCursor(text_cursor)
            text_cursor.endEditBlock()

    def _on_update_number_width(self, value=0):
        self.setViewportMargins(self._line_number_width(), 0, 0, 0)

    def _on_update_number_area(self, rect, y_value):
        if y_value:
            self._line_numbers.scroll(0, y_value)
        else:
            self._line_numbers.update(0, rect.y(), self._line_number_width(), rect.height())

        if rect.contains(self.viewport().rect()):
            self._update_number_width()

    def _on_cursor_position_changed(self):
        self._line_number_highlight()

    def _on_code_text_size_changed(self, value):
        self._set_text_size(value)

    def _on_activate(self):
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different classes to handle Qt settings
"""

from __future__ import print_function, division, absolute_import

import os
import sys

from Qt.QtCore import QSettings, QPoint, QSize
from Qt.QtWidgets import QMainWindow, QDockWidget, QComboBox, QCheckBox, QToolButton, QSpinBox, QLineEdit
from Qt.QtWidgets import QDoubleSpinBox

from tpDcc.libs.python import python, strings, settings


class QtSettings(QSettings, object):
    def __init__(self, filename, window=None, max_files=10,):
        super(QtSettings, self).__init__(filename, QSettings.IniFormat, window)

        self._max_files = max_files
        self._window = window
        if self._window:
            self._groups = [window.objectName(), 'RecentFiles']
        self._initialize()

    def has_setting(self, setting_name, setting_group=None):
        return bool(self.get(setting_name, setting_group=setting_group))

    def get(self, setting_name, default_value=None, setting_group=None, begin_group=None):
        """
        Returns the setting stored with the given name
        :param setting_name: str
        :param default_value: variant
        :param default_value: variant
        :param setting_group: str
        :return:
        """

        if setting_group:
            setting_name = '{}/{}'.format(setting_group, setting_name)

        if begin_group:
            self.beginGroup(begin_group)

        val = self.value(setting_name)
        if not val:
            if begin_group:
                self.endGroup()
            return default_value

        if python.is_string(val) and val.lower() in ['true', 'false']:
            if begin_group:
                self.endGroup()
            return strings.to_boolean(val)

        if begin_group:
            self.endGroup()

        return val

    def getw(self, setting_name, default_value=None):
        """
        Returns the window setting stored with the given name
        :param setting_name: str
        :param default_value: variant
        :return:
        """

        if self._window:
            val = self.value(self._window.objectName().upper() + '/' + setting_name)
            if not val:
                return default_value
        else:
            val = self.value(setting_name)
            if not val:
                return default_value

        return val

    def set(self, setting_name, setting_value, setting_group=None):
        """
        Stores a new settings with the given name and the given value
        If the given setting already exists, it will be overwrite
        :param setting_name: str, setting name we want store
        :param setting_value: variant, setting value we want to store
        """

        if setting_group:
            setting_name = '{}/{}'.format(setting_group, setting_name)

        self.setValue(setting_name, setting_value)

    def setw(self, setting_name, setting_value):
        """
        Stores a new window setting with the given name and the given value
        :param setting_name: str, setting name we want to store
        :param setting_value: variant, setting value we want to store
        """

        if self._window:
            self.setValue(self._window.objectName().upper() + '/' + setting_name, setting_value)
        else:
            self.setValue(setting_name, setting_value)

    def get_groups(self):
        """
        Returns the current preferences groups
        :return: list
        """

        return self._groups

    groups = property(get_groups)

    def add_group(self, group):
        """
        Add a group to the current preferences groups
        :param group: str, name of group to add
        :return: bool, True if the group was successfully added
        """

        if group not in self._groups:
            self._groups.append(group)
            return True

        return False

    def remove_group(self, group):
        """
        Remove a group from the preferences group
        :param group: str, name of group to remove
        :return: bool, True if the group was successfully removed
        """

        if group in self._groups:
            self._groups.remove(group)
            return True

        return False

    def window_keys(self):
        """
        Returns a list of all window keys to save for layouts or rebuilding last layout on launch
        :return: list
        """

        if self._window is None:
            return []

        result = [self._window.objectName()]
        for dock in self._window.findChildren(QDockWidget):
            dock_name = dock.objectName()
            result.append(str(dock_name))

        return result

    def prefs_keys(self):
        """
        Returns a list of preferences keys
        :return: list<str>, list of user prefs keys
        """

        results = list()
        self.beginGroup('Preferences')
        results = self.childKeys()
        self.endGroup()
        return results

    def get_layouts(self):
        """
        Returns a list of window layout names
        :return: list
        """

        layout_names = list()
        layout_keys = ['%s/geometry' % x for x in self.window_keys()]

        for k in self.allKeys():
            if 'geometry' in k:
                attrs = k.split('/geometry/')
                if len(attrs) > 1:
                    layout_names.append(str(attrs[-1]))

        return sorted(list(set(layout_names)))

    def save_layout(self, layout):
        """
        Save a named layout
        :param layout: str, layout name to save
        """

        sys.utils.logger.info('Saving Layout: "{}"'.format(layout))
        self.setValue(self._window.objectName() + '/geometry/%s' % layout, self._window.saveGeoemtry())
        self.setValue(self._window.objectName() + '/windowState/%s' % layout, self._window.saveState())

        for dock in self._window.findChildren(QDockWidget):
            dock_name = dock.objectName()
            self.setValue('%s/geometry/%s' % (dock_name, layout), dock.saveGeometry())

    def restore_layout(self, layout):
        """
        Restore a named layout
        :param layout: str, layout name to restores
        """

        sys.utils.logger.info('Restoring layout: "{}"'.format(layout))
        window_keys = self.window_keys()

        for widget_name in window_keys:
            key_name = '%s/geometry/%s' % (widget_name, layout)
            if widget_name != self._window.objectName():
                dock = self._window.findChildren(QDockWidget, widget_name)
                if dock:
                    if key_name in self.allKeys():
                        value = self.value(key_name)
                        dock[0].restoreGeometry(value)
            else:
                if key_name in self.allKeys():
                    value = self.value(key_name)
                    self._window.restoreGeometry(value)

                window_state = '%s/windowState/%s' % (widget_name, layout)
                if window_state in self.allKeys():
                    self._window.restoreState(self.value(window_state))

    def delete_layout(self, layout):
        """
        Delete a named layout
        :param layout: str, layout name to restore
        """

        sys.utils.logger.info('Deleting layout: "{}"'.format(layout))
        window_keys = self.window_keys()

        for widget_name in window_keys:
            key_name = '%s/geometry/%s' % (widget_name, layout)
            if key_name in self.allKeys():
                self.remove(key_name)

            if widget_name == self._window.objectName():
                window_state = '%s/windowState/%s' % (widget_name, layout)
                if window_state in self.allKeys():
                    self.remoev(window_state)

    def get_default_value(self, key, *groups):
        """
        Returns the default values for a group
        :param key: str, key to search for
        :return: variant, default value of key (None if not found)
        """

        if self.group():
            try:
                self.endGroup()
            except Exception:
                pass

        result = None

        if not groups:
            return None

        group_name = groups[0]
        for group in groups[1:]:
            group_name += '/%s' % group

        group_name += '/%s' % key
        group_name += '/%s' % 'default'

        if group_name in self.allKeys():
            result = self.value(group_name)

        return result

    def delete_file(self):
        """
        Delete the preferences file on disk
        """

        sys.utils.logger.info('Deleting Settings: "{}"'.format(self.fileName()))
        return os.remove(self.fileName())

    def get_recent_files(self):
        """
        Get a tuple of the most recent files
        """

        recent_files = list()
        cnt = self.beginReadArray('RecentFiles')
        for i in range(cnt):
            self.setArrayIndex(i)
            fn = self.value('file')
            recent_files.append(fn)
        self.endArray()

        return tuple(recent_files)

    def add_recent_file(self, filename):
        """
        Adds a recent file to the stack
        :param filename: str, file name to add
        """

        recent_files = self.get_recent_files()
        if filename in recent_files:
            recent_files = tuple(x for x in recent_files if x != filename)

        recent_files = recent_files + (filename,)
        self.beginWriteArray('RecentFiles')
        for i in range(len(recent_files)):
            self.setArrayIndex(i)
            self.setValue('file', recent_files[i])
        self.endArray()

    def clear_recent_files(self):
        self.remove('RecentFiles')

    def _initialize(self):
        if self._window:
            window_name = self._window.objectName().upper()
            if window_name not in self.childGroups():
                if self._window is not None:
                    self.setValue(window_name + '/geometry/default', self._window.saveGeometry())

                    if isinstance(self._window, QMainWindow):
                        self.setValue(window_name + '/windowState/default', self._window.saveState())

            if 'RecentFiles' not in self.childGroups():
                self.beginWriteArray('RecentFiles', 0)
                self.endArray()

            while self.group():
                self.endGroup()


class QtIniSettings(settings.INISettings, object):
    def __init__(self, filename):
        super(QtIniSettings, self).__init__(filename)

    def export_widget(self, option, widget):
        """
        Serializes a widget's state to an option, value pair Sets the value with an associated key in
        the section at the top of the section stack.

        :param str option:   The key to use in storing the value.
        :param str widget:   The QWidget to extract value to store.
        """

        if isinstance(widget, QComboBox):
            self.export_option(option, widget.currentIndex())
        elif isinstance(widget, QCheckBox):
            self.export_option(option, widget.isChecked())
        elif isinstance(widget, QToolButton):
            if widget.isCheckable():
                self.export_option(option, widget.isChecked())
        elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
            self.export_option(option, widget.value())
        elif isinstance(widget, QLineEdit):
            self.export_option(option, widget.text())
        elif isinstance(widget, (QSize, QPoint)):
            self.export_option(option, widget.toTuple())
        else:
            assert False, "Unknown control type"

    def import_widget(self, option, widget):
        """
        Serializes a widget's state to an option, value pair Sets the value with an associated key in the
        section at the top of the section stack.

        :param str option:   The key to use in storing the value.
        :param str widget:   The QWidget to extract value to store.
        """

        if isinstance(widget, QComboBox):
            d = 0 if (widget.currentIndex() < 0) else widget.currentIndex()
            v = self.import_option(option, d)
            widget.setCurrentIndex(v)
        elif isinstance(widget, QCheckBox):
            d = widget.isChecked()
            v = self.import_option(option, d)
            widget.setChecked(v)
        elif isinstance(widget, QToolButton):
            if widget.isCheckable():
                d = widget.isChecked()
                v = self.import_option(option, d)
                widget.setChecked(v)
        elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
            d = widget.value()
            v = self.import_option(option, d)
            widget.setValue(v)
        elif isinstance(widget, QLineEdit):
            d = widget.text()
            v = self.import_option(option, d)
            widget.setText(v)
        elif isinstance(widget, QSize):
            d = widget.toTuple()
            v = self.import_option(option, d, True)
            widget.setWidth(v[0])
            widget.setHeight(v[1])
        elif isinstance(widget, QPoint):
            d = widget.toTuple()
            v = self.import_option(option, d, True)
            widget.setX(v[0])
            widget.setY(v[1])
        else:
            assert False, "Unknown control type"



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom Qt widgets related with data management
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal
from Qt.QtWidgets import QSizePolicy

from tpDcc.managers import resources
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, buttons


class SaveFileWidet(base.DirectoryWidget):

    fileChanged = Signal()

    TIP = ''

    def __init__(self, parent=None):

        self._data_class = None

        super(SaveFileWidet, self).__init__(parent=parent)

        self._tip = self.TIP
        if self._tip:
            self._create_tip()

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def get_main_layout(self):
        main_layout = layouts.HorizontalLayout()
        main_layout.setAlignment(Qt.AlignTop)

        return main_layout

    def ui(self):
        super(SaveFileWidet, self).ui()

        self.setContentsMargins(1, 1, 1, 1)

        self._save_button = buttons.BaseButton('Save', parent=self)
        self._load_button = buttons.BaseButton('Open', parent=self)
        self._save_button.setMaximumWidth(100)
        self._load_button.setMaximumWidth(100)
        self._save_button.setMinimumWidth(60)
        self._load_button.setMinimumWidth(60)
        self._save_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self._load_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self._save_button.setIcon(resources.icon('save'))
        self._load_button.setIcon(resources.icon('folder'))

        self.main_layout.addWidget(self._save_button)
        self.main_layout.addWidget(self._load_button)

    def setup_signals(self):
        self._save_button.clicked.connect(self._on_save)
        self._load_button.clicked.connecxt(self._on_load)

    def set_directory(self, directory, data_class=None):
        super(SaveFileWidet, self).set_directory(directory)

        if data_class:
            self._data_class = data_class
        if not data_class and self._data_class:
            self._data_class.set_directory(self._directory)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_tip(self, value):
        self._tip = value
        if self._tip:
            self._create_tip()

    def set_no_save(self):
        self._save_button.setEnabled(False)

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _create_tip(self):
        self.setToolTip(self._tip)

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_save(self):
        self.fileChanged.emit()

    def _on_load(self):
        pass



from Qt.QtCore import *

import tpDcc
from tpDcc.libs.qt.core import mixin
from tpDcc.libs.qt.widgets import layouts, label, browser, dividers


class BrowserExample(tpDcc.Window, mixin.FieldMixin):
    def __init__(self):
        super(BrowserExample, self).__init__(title='Example for Browsers')

    def ui(self):
        super(BrowserExample, self).ui()

        browser1 = browser.ClickBrowserFileButton('Browse File PushButton', parent=self).primary()
        browser2 = browser.ClickBrowserFolderButton('Browse Folder PushButton', parent=self)
        browser2.setIcon(tpDcc.ResourcesMgr().icon('upload'))
        browser3 = browser.ClickBrowserFileButton('Browse Multi Files', multiple=True, parent=self).primary()
        browser4 = browser.ClickBrowserFileToolButton().huge()
        lbl4 = label.BaseLabel()
        lbl4._set_elide_mode(Qt.ElideMiddle)
        browser4.fileChanged.connect(lbl4.setText)
        browser5 = browser.ClickBrowserFolderToolButton().huge()
        lbl5 = label.BaseLabel()
        lbl5._set_elide_mode(Qt.ElideMiddle)
        browser5.folderChanged.connect(lbl5.setText)
        browser6 = browser.DragFileButton('Click or drag file here')
        browser6.image = 'attachment'
        lbl6 = label.BaseLabel()
        lbl6._set_elide_mode(Qt.ElideMiddle)
        browser6.fileChanged.connect(lbl6.setText)
        browser7 = browser.DragFolderButton()
        lbl7 = label.BaseLabel()
        lbl7._set_elide_mode(Qt.ElideMiddle)
        browser7.folderChanged.connect(lbl7.setText)
        browser8 = browser.DragFileButton('Click or drag media file here')
        browser8.image('video_file')
        browser8.filters = ['.mov', '.mp4']
        lbl8 = label.BaseLabel()
        lbl8._set_elide_mode(Qt.ElideRight)
        self.register_field('current_file', '')
        self.bind('current_file', browser8, 'path', signal='fileChanged')
        self.bind('current_file', lbl8, 'text')

        lyt1 = layouts.HorizontalLayout()
        lyt1.addWidget(browser1)
        lyt1.addWidget(browser2)
        lyt1.addWidget(browser3)

        lyt2 = layouts.HorizontalLayout()
        lyt2.addWidget(lbl4)
        lyt2.addWidget(browser4)
        lyt2.addWidget(lbl5)
        lyt2.addWidget(browser5)

        lyt3 = layouts.GridLayout()
        lyt3.addWidget(browser6, 2, 0)
        lyt3.addWidget(lbl6, 3, 0)
        lyt3.addWidget(browser7, 2, 1)
        lyt3.addWidget(lbl7, 3, 1)

        self.main_layout.addWidget(dividers.Divider('ClickBrowserFileButton'))
        self.main_layout.addLayout(lyt1)
        self.main_layout.addWidget(dividers.Divider('ClickBrowserToolButton'))
        self.main_layout.addLayout(lyt2)
        self.main_layout.addWidget(dividers.Divider('DragBrowserToolButton'))
        self.main_layout.addLayout(lyt3)
        self.main_layout.addWidget(dividers.Divider('data bind'))
        self.main_layout.addWidget(browser8)
        self.main_layout.addWidget(lbl8)
        self.main_layout.addStretch()


def run():
    return BrowserExample()



from Qt.QtCore import *

import tpDcc
from tpDcc.libs.qt.core import mixin, menu
from tpDcc.libs.qt.widgets import layouts, label, buttons, lineedit, dividers, message, combobox


class LineEditExample(tpDcc.Window, mixin.FieldMixin):
    def __init__(self):
        super(LineEditExample, self).__init__(title='Example for Line Edits')

    def ui(self):
        super(LineEditExample, self).ui()

        size_lyt = layouts.HorizontalLayout()
        line_edit_small = lineedit.BaseLineEdit().small()
        line_edit_small.setPlaceholderText('small size')
        line_edit_medium = lineedit.BaseLineEdit().medium()
        line_edit_medium.setPlaceholderText('default size')
        line_edit_large = lineedit.BaseLineEdit().large()
        line_edit_large.setPlaceholderText('large size')
        size_lyt.addWidget(line_edit_small)
        size_lyt.addWidget(line_edit_medium)
        size_lyt.addWidget(line_edit_large)

        line_edit_tool_btn = lineedit.BaseLineEdit(text='BaseToolButton')
        line_edit_tool_btn.set_prefix_widget(buttons.BaseToolButton().image('user').icon_only())
        line_edit_label = lineedit.BaseLineEdit(text='BaseLabel')
        line_lbl = label.BaseLabel('User').mark().secondary()
        line_lbl.setAlignment(Qt.AlignCenter)
        line_lbl.setFixedWidth(80)
        line_edit_label.set_prefix_widget(line_lbl)
        line_edit_btn = lineedit.BaseLineEdit('BaseButton')
        push_btn = buttons.BaseButton('Go').primary()
        line_edit_btn.set_suffix_widget(push_btn)
        search_engine_line = lineedit.BaseLineEdit().search_engine().large()
        search_engine_line.returnPressed.connect(self._on_search)
        line_edit_options = lineedit.BaseLineEdit()
        combo_box = combobox.BaseComboBox()
        option_menu = menu.BaseMenu()
        option_menu.set_separator('|')
        option_menu.set_data([r'http://', r'https://'])
        combo_box.set_menu(option_menu)
        combo_box.set_value('http://')
        combo_box.setFixedWidth(90)
        line_edit_options.set_prefix_widget(combo_box)

        self.main_layout.addWidget(dividers.Divider('different size'))
        self.main_layout.addLayout(size_lyt)
        self.main_layout.addWidget(dividers.Divider('custom prefix and suffix widget'))
        self.main_layout.addWidget(line_edit_tool_btn)
        self.main_layout.addWidget(line_edit_label)
        self.main_layout.addWidget(line_edit_btn)
        self.main_layout.addWidget(dividers.Divider('preset'))
        self.main_layout.addWidget(label.BaseLabel('error'))
        self.main_layout.addWidget(lineedit.BaseLineEdit('waring: file d:/ddd/ccc.jpg not exists.').error())
        self.main_layout.addWidget(label.BaseLabel('search'))
        self.main_layout.addWidget(lineedit.BaseLineEdit().search().small())
        self.main_layout.addWidget(label.BaseLabel('search engine'))
        self.main_layout.addWidget(search_engine_line)
        self.main_layout.addWidget(label.BaseLabel('file'))
        self.main_layout.addWidget(lineedit.BaseLineEdit().file().small())
        self.main_layout.addWidget(label.BaseLabel('folder'))
        self.main_layout.addWidget(lineedit.BaseLineEdit().folder().small())
        self.main_layout.addWidget(label.BaseLabel('options'))
        self.main_layout.addWidget(line_edit_options)
        self.main_layout.addStretch()

    def _on_search(self):
        message.PopupMessage.info('Searching ....', parent=self)


def run():
    return LineEditExample()
